# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Generator
from dataclasses import dataclass
from typing import Optional

import numpy as np

from braket.annealing import ProblemType
from braket.task_result import AdditionalMetadata, AnnealingTaskResult, TaskMetadata


@dataclass
class AnnealingQuantumTaskResult:
    """Result of an annealing problem quantum task execution. This class is intended
    to be initialized by a QuantumTask class.

    Args:
        record_array (np.recarray): numpy array with keys 'solution' (np.ndarray)
            where row is solution, column is value of the variable, 'solution_count' (numpy.ndarray)
            the number of times the solutions occurred, and 'value' (numpy.ndarray) the
            output or energy of the solutions.
        variable_count (int): the number of variables
        problem_type (ProblemType): the type of annealing problem
        task_metadata (TaskMetadata): Quantum task metadata.
        additional_metadata (AdditionalMetadata): Additional metadata about the quantum task
    """

    record_array: np.recarray
    variable_count: int
    problem_type: ProblemType
    task_metadata: TaskMetadata
    additional_metadata: AdditionalMetadata

    def data(
        self,
        selected_fields: Optional[list[str]] = None,
        sorted_by: str = "value",
        reverse: bool = False,
    ) -> Generator[tuple]:
        """Yields the data in record_array

        Args:
            selected_fields (Optional[list[str]]): selected fields to return.
                Options are 'solution', 'value', and 'solution_count'. Default is None.
            sorted_by (str): Sorts the data by this field.
                Options are 'solution', 'value', and 'solution_count'. Default is 'value'.
            reverse (bool): If True, returns the data in reverse order. Default is False.

        Yields:
            Generator[tuple]: data in record_array
        """
        if selected_fields is None:
            selected_fields = ["solution", "value", "solution_count"]

        if sorted_by is None:
            order = np.arange(len(self.record_array))
        else:
            order = np.argsort(self.record_array[sorted_by])

        if reverse:
            order = np.flip(order)

        for i in order:
            yield tuple(self.record_array[field][i] for field in selected_fields)

    def __eq__(self, other: AnnealingQuantumTaskResult) -> bool:
        if isinstance(other, AnnealingQuantumTaskResult):
            # __eq__ on numpy arrays results in an array of booleans and therefore can't use
            # the default dataclass __eq__ implementation. Override equals to check if all
            # elements in the array are equal.
            self_fields = (
                self.variable_count,
                self.problem_type,
                self.task_metadata,
                self.additional_metadata,
            )
            other_fields = (
                other.variable_count,
                other.problem_type,
                other.task_metadata,
                other.additional_metadata,
            )
            return (self.record_array == other.record_array).all() and self_fields == other_fields
        return NotImplemented

    @staticmethod
    def from_object(result: AnnealingTaskResult) -> AnnealingQuantumTaskResult:
        """Create AnnealingQuantumTaskResult from AnnealingTaskResult object

        Args:
            result (AnnealingTaskResult): AnnealingTaskResult object

        Returns:
            AnnealingQuantumTaskResult: An AnnealingQuantumTaskResult based on the
            given result object
        """
        return AnnealingQuantumTaskResult._from_object(result)

    @staticmethod
    def from_string(result: str) -> AnnealingQuantumTaskResult:
        """Create AnnealingQuantumTaskResult from string

        Args:
            result (str): JSON object string

        Returns:
            AnnealingQuantumTaskResult: An AnnealingQuantumTaskResult based on the given string
        """
        return AnnealingQuantumTaskResult._from_object(AnnealingTaskResult.parse_raw(result))

    @classmethod
    def _from_object(cls, result: AnnealingTaskResult) -> AnnealingQuantumTaskResult:
        solutions = np.asarray(result.solutions, dtype=int)
        values = np.asarray(result.values, dtype=float)
        if not result.solutionCounts:
            solution_counts = np.ones(len(solutions), dtype=int)
        else:
            solution_counts = np.asarray(result.solutionCounts, dtype=int)
        record_array = AnnealingQuantumTaskResult._create_record_array(
            solutions, solution_counts, values
        )
        variable_count = result.variableCount
        problem_type = ProblemType[result.additionalMetadata.action.type.value]
        task_metadata = result.taskMetadata
        additional_metadata = result.additionalMetadata
        return cls(
            record_array=record_array,
            variable_count=variable_count,
            problem_type=problem_type,
            task_metadata=task_metadata,
            additional_metadata=additional_metadata,
        )

    @staticmethod
    def _create_record_array(
        solutions: np.ndarray, solution_counts: np.ndarray, values: np.ndarray
    ) -> np.recarray:
        """Create a solutions record for AnnealingQuantumTaskResult

        Args:
            solutions (np.ndarray): row is solution, column is value of the variable
            solution_counts (np.ndarray): list of number of times the solutions occurred
            values (np.ndarray): list of the output or energy of the solutions

        Returns:
            np.recarray: A record array for solutions, value, and solution_count.
        """
        num_solutions, variable_count = solutions.shape
        datatypes = [
            ("solution", solutions.dtype, (variable_count,)),
            ("value", values.dtype),
            ("solution_count", solution_counts.dtype),
        ]

        record = np.rec.array(np.zeros(num_solutions, dtype=datatypes))
        record["solution"] = solutions
        record["value"] = values
        record["solution_count"] = solution_counts
        return record



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from typing import Optional, Union

import numpy as np
from sympy import Expr, Number

from braket.circuits import Circuit, Instruction
from braket.circuits.gates import Unitary
from braket.circuits.measure import Measure
from braket.circuits.noises import Kraus
from braket.circuits.translations import (
    BRAKET_GATES,
    braket_result_to_result_type,
    one_prob_noise_map,
)
from braket.default_simulator.openqasm.program_context import AbstractProgramContext
from braket.ir.jaqcd.program_v1 import Results
from braket.parametric import FreeParameterExpression


class BraketProgramContext(AbstractProgramContext):
    def __init__(self, circuit: Optional[Circuit] = None):
        """Inits a `BraketProgramContext`.

        Args:
            circuit (Optional[Circuit]): A partially-built circuit to continue building with this
                context. Default: None.
        """
        super().__init__()
        self._circuit = circuit or Circuit()

    @property
    def circuit(self) -> Circuit:
        """The circuit being built in this context."""
        return self._circuit

    def is_builtin_gate(self, name: str) -> bool:
        """Whether the gate is currently in scope as a built-in Braket gate.

        Args:
            name (str): name of the built-in Braket gate

        Returns:
            bool: return TRUE if it is a built-in gate else FALSE.
        """
        user_defined_gate = self.is_user_defined_gate(name)
        return name in BRAKET_GATES and not user_defined_gate

    def add_phase_instruction(self, target: tuple[int], phase_value: float) -> None:
        """Add a global phase to the circuit.

        Args:
            target (tuple[int]): Unused
            phase_value (float): The phase value to be applied
        """
        instruction = Instruction(BRAKET_GATES["gphase"](phase_value))
        self._circuit.add_instruction(instruction)

    def add_gate_instruction(
        self, gate_name: str, target: tuple[int], *params, ctrl_modifiers: list[int], power: float
    ) -> None:
        """Add Braket gate to the circuit.

        Args:
            gate_name (str): name of the built-in Braket gate.
            target (tuple[int]): control_qubits + target_qubits.
            ctrl_modifiers (list[int]): Quantum state on which to control the
                operation. Must be a binary sequence of same length as number of qubits in
                `control-qubits` in target. For example "0101", [0, 1, 0, 1], 5 all represent
                controlling on qubits 0 and 2 being in the \\|0⟩ state and qubits 1 and 3 being
                in the \\|1⟩ state.
            power(float): Integer or fractional power to raise the gate to.
        """
        target_qubits = target[len(ctrl_modifiers) :]
        control_qubits = target[: len(ctrl_modifiers)]
        instruction = Instruction(
            BRAKET_GATES[gate_name](*params[0]),
            target=target_qubits,
            control=control_qubits,
            control_state=ctrl_modifiers,
            power=power,
        )
        self._circuit.add_instruction(instruction)

    def add_custom_unitary(
        self,
        unitary: np.ndarray,
        target: tuple[int],
    ) -> None:
        """Add a custom Unitary instruction to the circuit

        Args:
            unitary (np.ndarray): unitary matrix
            target (tuple[int]): control_qubits + target_qubits
        """
        instruction = Instruction(Unitary(unitary), target)
        self._circuit.add_instruction(instruction)

    def add_noise_instruction(
        self, noise_instruction: str, target: list[int], probabilities: list[float]
    ) -> None:
        """Method to add a noise instruction to the circuit

        Args:
            noise_instruction (str): The name of the noise operation
            target (list[int]): The target qubit or qubits to which the noise operation is applied.
            probabilities (list[float]): The probabilities associated with each possible outcome
                of the noise operation.
        """
        instruction = Instruction(
            one_prob_noise_map[noise_instruction](*probabilities), target=target
        )
        self._circuit.add_instruction(instruction)

    def add_kraus_instruction(self, matrices: list[np.ndarray], target: list[int]) -> None:
        """Method to add a Kraus instruction to the circuit

        Args:
            matrices (list[ndarray]): The matrices defining the Kraus operation
            target (list[int]): The target qubit or qubits to which the Kraus operation is applied.
        """
        instruction = Instruction(Kraus(matrices), target)
        self._circuit.add_instruction(instruction)

    def add_result(self, result: Results) -> None:
        """Abstract method to add result type to the circuit

        Args:
            result (Results): The result object representing the measurement results
        """
        self._circuit.add_result_type(braket_result_to_result_type(result))

    def handle_parameter_value(
        self, value: Union[float, Expr]
    ) -> Union[float, FreeParameterExpression]:
        """Convert parameter value to required format.

        Args:
            value (Union[float, Expr]): Value of the parameter

        Returns:
            Union[float, FreeParameterExpression]: Return the value directly if numeric,
            otherwise wraps the symbolic expression as a `FreeParameterExpression`.
        """
        if isinstance(value, Expr):
            evaluated_value = value.evalf()
            if isinstance(evaluated_value, Number):
                return evaluated_value
            return FreeParameterExpression(evaluated_value)
        return value

    def add_measure(self, target: tuple[int]) -> None:
        """Add a measure instruction to the circuit

        Args:
            target (tuple[int]): the target qubits to be measured.
        """
        instruction = Instruction(Measure(), list(target))
        self._circuit.add_instruction(instruction)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import datetime
import tempfile
import time
from collections import defaultdict
from dataclasses import asdict
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from braket.aws import AwsSession
from braket.jobs import Framework, retrieve_image
from braket.jobs.config import (
    CheckpointConfig,
    InstanceConfig,
    OutputDataConfig,
    S3DataSourceConfig,
    StoppingCondition,
)
from braket.jobs.quantum_job_creation import (
    _exclude_nones_factory,
    _generate_default_job_name,
    _process_input_data,
    _process_local_source_module,
    _process_s3_source_module,
    _tar_and_upload_to_code_location,
    _validate_entry_point,
    prepare_quantum_job,
)


@pytest.fixture
def aws_session():
    _aws_session = Mock(spec=AwsSession)
    _aws_session.default_bucket.return_value = "default-bucket-name"
    _aws_session.get_default_jobs_role.return_value = "default-role-arn"
    _aws_session.region = "us-east-1"
    return _aws_session


@pytest.fixture
def entry_point():
    return "test-source-dir.entry_point:func"


@pytest.fixture
def bucket():
    return "braket-region-id"


@pytest.fixture
def tags():
    return {"tag-key": "tag-value"}


@pytest.fixture(
    params=[
        None,
        "aws.location/amazon-braket-custom-jobs:tag.1.2.3",
        "other.uri/amazon-braket-custom-name:tag",
        "other.uri/custom-non-managed:tag",
        "other-custom-format.com",
    ]
)
def image_uri(request):
    return request.param


@pytest.fixture(params=["given_job_name", "default_job_name"])
def job_name(request):
    if request.param == "given_job_name":
        return "test-job-name"


@pytest.fixture
def s3_prefix(job_name):
    return f"{job_name}/non-default"


@pytest.fixture(params=["local_source", "s3_source"])
def source_module(request, bucket):
    if request.param == "local_source":
        return "test-source-module"
    elif request.param == "s3_source":
        return AwsSession.construct_s3_uri(bucket, "test-source-prefix", "source.tar.gz")


@pytest.fixture
def code_location(bucket, s3_prefix):
    return AwsSession.construct_s3_uri(bucket, s3_prefix, "script")


@pytest.fixture
def role_arn():
    return "arn:aws:iam::0000000000:role/AmazonBraketInternalSLR"


@pytest.fixture
def device():
    return "arn:aws:braket:::device/qpu/test/device-name"


@pytest.fixture
def hyperparameters():
    return {
        "param": "value",
        "other-param": 100,
    }


@pytest.fixture(params=["dict", "local"])
def input_data(request, bucket):
    if request.param == "dict":
        return {
            "s3_input": f"s3://{bucket}/data/prefix",
            "local_input": "local/prefix",
            "config_input": S3DataSourceConfig(f"s3://{bucket}/config/prefix"),
        }
    elif request.param == "local":
        return "local/prefix"


@pytest.fixture
def instance_config():
    return InstanceConfig(
        instanceType="ml.m5.large",
        volumeSizeInGb=1,
    )


@pytest.fixture(params=[False, True])
def data_parallel(request):
    return request.param


@pytest.fixture
def distribution(data_parallel):
    if data_parallel:
        return "data_parallel"
    return None


@pytest.fixture
def stopping_condition():
    return StoppingCondition(
        maxRuntimeInSeconds=1200,
    )


@pytest.fixture
def output_data_config(bucket, s3_prefix):
    return OutputDataConfig(
        s3Path=AwsSession.construct_s3_uri(bucket, s3_prefix, "output"),
    )


@pytest.fixture
def checkpoint_config(bucket, s3_prefix):
    return CheckpointConfig(
        localPath="/opt/omega/checkpoints",
        s3Uri=AwsSession.construct_s3_uri(bucket, s3_prefix, "checkpoints"),
    )


@pytest.fixture
def reservation_arn():
    return "arn:aws:braket:us-west-2:123456789123:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"


@pytest.fixture
def generate_get_job_response():
    def _get_job_response(**kwargs):
        response = {
            "ResponseMetadata": {
                "RequestId": "d223b1a0-ee5c-4c75-afa7-3c29d5338b62",
                "HTTPStatusCode": 200,
            },
            "algorithmSpecification": {
                "scriptModeConfig": {
                    "entryPoint": "my_file:start_here",
                    "s3Uri": "s3://amazon-braket-jobs/job-path/my_file.py",
                }
            },
            "checkpointConfig": {
                "localPath": "/opt/omega/checkpoints",
                "s3Uri": "s3://amazon-braket-jobs/job-path/checkpoints",
            },
            "createdAt": datetime.datetime(2021, 6, 28, 21, 4, 51),
            "deviceConfig": {
                "device": "arn:aws:braket:::device/qpu/rigetti/Aspen-10",
            },
            "hyperParameters": {
                "foo": "bar",
            },
            "inputDataConfig": [
                {
                    "channelName": "training_input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://amazon-braket-jobs/job-path/input",
                        }
                    },
                }
            ],
            "instanceConfig": {
                "instanceType": "ml.m5.large",
                "volumeSizeInGb": 1,
            },
            "jobArn": "arn:aws:braket:us-west-2:875981177017:job/job-test-20210628140446",
            "jobName": "job-test-20210628140446",
            "outputDataConfig": {"s3Path": "s3://amazon-braket-jobs/job-path/data"},
            "roleArn": "arn:aws:iam::875981177017:role/AmazonBraketJobRole",
            "status": "RUNNING",
            "stoppingCondition": {"maxRuntimeInSeconds": 1200},
        }
        response.update(kwargs)

        return response

    return _get_job_response


@pytest.fixture(params=["fixtures", "defaults", "nones"])
def create_job_args(
    request,
    aws_session,
    entry_point,
    image_uri,
    source_module,
    job_name,
    code_location,
    role_arn,
    device,
    hyperparameters,
    input_data,
    instance_config,
    distribution,
    stopping_condition,
    output_data_config,
    checkpoint_config,
    tags,
    reservation_arn,
):
    if request.param == "fixtures":
        return dict(
            (key, value)
            for key, value in {
                "device": device,
                "source_module": source_module,
                "entry_point": entry_point,
                "image_uri": image_uri,
                "job_name": job_name,
                "code_location": code_location,
                "role_arn": role_arn,
                "hyperparameters": hyperparameters,
                "input_data": input_data,
                "instance_config": instance_config,
                "distribution": distribution,
                "stopping_condition": stopping_condition,
                "output_data_config": output_data_config,
                "checkpoint_config": checkpoint_config,
                "aws_session": aws_session,
                "tags": tags,
                "reservation_arn": reservation_arn,
            }.items()
            if value is not None
        )
    elif request.param == "defaults":
        return {
            "device": device,
            "source_module": source_module,
            "entry_point": entry_point,
            "aws_session": aws_session,
        }
    elif request.param == "nones":
        return defaultdict(
            lambda: None,
            device=device,
            source_module=source_module,
            entry_point=entry_point,
            aws_session=aws_session,
        )


@patch("tarfile.TarFile.add")
@patch("importlib.util.find_spec")
@patch("braket.jobs.quantum_job_creation.Path")
@patch("time.time")
def test_create_job(
    mock_time,
    mock_path,
    mock_findspec,
    mock_tarfile,
    aws_session,
    source_module,
    create_job_args,
):
    mock_path.return_value.resolve.return_value.parent = "parent_dir"
    mock_path.return_value.resolve.return_value.stem = source_module
    mock_path.return_value.name = "file_name"
    mock_time.return_value = datetime.datetime.now().timestamp()
    expected_kwargs = _translate_creation_args(create_job_args)
    result_kwargs = prepare_quantum_job(**create_job_args)
    assert expected_kwargs == result_kwargs


def _translate_creation_args(create_job_args):
    aws_session = create_job_args["aws_session"]
    create_job_args = defaultdict(lambda: None, **create_job_args)
    image_uri = create_job_args["image_uri"]
    job_name = create_job_args["job_name"] or _generate_default_job_name(image_uri)
    default_bucket = aws_session.default_bucket()
    timestamp = str(int(time.time() * 1000))
    code_location = create_job_args["code_location"] or AwsSession.construct_s3_uri(
        default_bucket, "jobs", job_name, timestamp, "script"
    )
    role_arn = create_job_args["role_arn"] or aws_session.get_default_jobs_role()
    device = create_job_args["device"]
    hyperparameters = create_job_args["hyperparameters"] or {}
    hyperparameters = {str(key): str(value) for key, value in hyperparameters.items()}
    input_data = create_job_args["input_data"] or {}
    instance_config = create_job_args["instance_config"] or InstanceConfig()
    reservation_arn = create_job_args["reservation_arn"]
    if create_job_args["distribution"] == "data_parallel":
        distributed_hyperparams = {
            "sagemaker_distributed_dataparallel_enabled": "true",
            "sagemaker_instance_type": instance_config.instanceType,
        }
        hyperparameters.update(distributed_hyperparams)
    output_data_config = create_job_args["output_data_config"] or OutputDataConfig(
        s3Path=AwsSession.construct_s3_uri(default_bucket, "jobs", job_name, timestamp, "data")
    )
    stopping_condition = create_job_args["stopping_condition"] or StoppingCondition()
    checkpoint_config = create_job_args["checkpoint_config"] or CheckpointConfig(
        s3Uri=AwsSession.construct_s3_uri(
            default_bucket, "jobs", job_name, timestamp, "checkpoints"
        )
    )
    entry_point = create_job_args["entry_point"]
    source_module = create_job_args["source_module"]
    if not AwsSession.is_s3_uri(source_module):
        entry_point = entry_point or Path(source_module).stem
    algorithm_specification = {
        "scriptModeConfig": {
            "entryPoint": entry_point,
            "s3Uri": f"{code_location}/source.tar.gz",
            "compressionType": "GZIP",
        }
    }
    image_uri = image_uri or retrieve_image(Framework.BASE, aws_session.region)
    algorithm_specification["containerImage"] = {"uri": image_uri}
    tags = create_job_args.get("tags", {})

    test_kwargs = {
        "jobName": job_name,
        "roleArn": role_arn,
        "algorithmSpecification": algorithm_specification,
        "inputDataConfig": _process_input_data(input_data, job_name, aws_session, timestamp),
        "instanceConfig": asdict(instance_config),
        "outputDataConfig": asdict(output_data_config, dict_factory=_exclude_nones_factory),
        "checkpointConfig": asdict(checkpoint_config),
        "deviceConfig": {"device": device},
        "hyperParameters": hyperparameters,
        "stoppingCondition": asdict(stopping_condition),
        "tags": tags,
    }

    if reservation_arn:
        test_kwargs.update(
            {
                "associations": [
                    {
                        "arn": reservation_arn,
                        "type": "RESERVATION_TIME_WINDOW_ARN",
                    }
                ]
            }
        )

    return test_kwargs


@patch("time.time")
def test_generate_default_job_name(mock_time, image_uri):
    job_type_mapping = {
        None: "-default",
        "aws.location/amazon-braket-custom-jobs:tag.1.2.3": "-custom",
        "other.uri/amazon-braket-custom-name:tag": "-custom-name",
        "other.uri/custom-non-managed:tag": "",
        "other-custom-format.com": "",
    }
    job_type = job_type_mapping[image_uri]
    mock_time.return_value = datetime.datetime.now().timestamp()
    timestamp = str(int(time.time() * 1000))
    assert _generate_default_job_name(image_uri) == f"braket-job{job_type}-{timestamp}"
    assert _generate_default_job_name(image_uri, timestamp="ts") == f"braket-job{job_type}-ts"


@pytest.mark.parametrize(
    "source_module",
    (
        "s3://bucket/source_module.tar.gz",
        "s3://bucket/SOURCE_MODULE.TAR.GZ",
    ),
)
def test_process_s3_source_module(source_module, aws_session):
    _process_s3_source_module(source_module, "entry_point", aws_session, "code_location")
    aws_session.copy_s3_object.assert_called_with(source_module, "code_location/source.tar.gz")


def test_process_s3_source_module_not_tar_gz(aws_session):
    must_be_tar_gz = (
        "If source_module is an S3 URI, it must point to a tar.gz file. "
        "Not a valid S3 URI for parameter `source_module`: s3://bucket/source_module"
    )
    with pytest.raises(ValueError, match=must_be_tar_gz):
        _process_s3_source_module(
            "s3://bucket/source_module", "entry_point", aws_session, "code_location"
        )


def test_process_s3_source_module_no_entry_point(aws_session):
    entry_point_required = "If source_module is an S3 URI, entry_point must be provided."
    with pytest.raises(ValueError, match=entry_point_required):
        _process_s3_source_module("s3://bucket/source_module", None, aws_session, "code_location")


@patch("braket.jobs.quantum_job_creation._tar_and_upload_to_code_location")
@patch("braket.jobs.quantum_job_creation._validate_entry_point")
def test_process_local_source_module(validate_mock, tar_and_upload_mock, aws_session):
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module = Path(temp_dir, "source_module")
        source_module.touch()

        _process_local_source_module(
            str(source_module), "entry_point", aws_session, "code_location"
        )

        source_module_abs_path = Path(temp_dir, "source_module").resolve()
        validate_mock.assert_called_with(source_module_abs_path, "entry_point")
        tar_and_upload_mock.assert_called_with(source_module_abs_path, aws_session, "code_location")


def test_process_local_source_module_not_found(aws_session):
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module = str(Path(temp_dir, "source_module").as_posix())
        source_module_not_found = f"Source module not found: {source_module}"
        with pytest.raises(ValueError, match=source_module_not_found):
            _process_local_source_module(source_module, "entry_point", aws_session, "code_location")


def test_validate_entry_point_default_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module_path = Path(temp_dir, "source_module.py")
        source_module_path.touch()
        # import source_module
        _validate_entry_point(source_module_path, "source_module")
        # from source_module import func
        _validate_entry_point(source_module_path, "source_module:func")
        # import .
        _validate_entry_point(source_module_path, ".")
        # from . import func
        _validate_entry_point(source_module_path, ".:func")


def test_validate_entry_point_default_directory():
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module_path = Path(temp_dir, "source_module")
        source_module_path.mkdir()
        # import source_module
        _validate_entry_point(source_module_path, "source_module")
        # from source_module import func
        _validate_entry_point(source_module_path, "source_module:func")
        # import .
        _validate_entry_point(source_module_path, ".")
        # from . import func
        _validate_entry_point(source_module_path, ".:func")


def test_validate_entry_point_submodule_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module_path = Path(temp_dir, "source_module")
        source_module_path.mkdir()
        Path(source_module_path, "submodule.py").touch()
        # from source_module import submodule
        _validate_entry_point(source_module_path, "source_module.submodule")
        # from source_module.submodule import func
        _validate_entry_point(source_module_path, "source_module.submodule:func")
        # from . import submodule
        _validate_entry_point(source_module_path, ".submodule")
        # from .submodule import func
        _validate_entry_point(source_module_path, ".submodule:func")


def test_validate_entry_point_submodule_init():
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module_path = Path(temp_dir, "source_module")
        source_module_path.mkdir()
        Path(source_module_path, "submodule.py").touch()
        with open(str(Path(source_module_path, "__init__.py")), "w") as f:
            f.write("from . import submodule as renamed")
        # from source_module import renamed
        _validate_entry_point(source_module_path, "source_module:renamed")
        # from . import renamed
        _validate_entry_point(source_module_path, ".:renamed")


def test_validate_entry_point_source_module_not_found():
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module_path = Path(temp_dir, "source_module")
        source_module_path.mkdir()
        Path(source_module_path, "submodule.py").touch()

        # catches ModuleNotFoundError
        module_not_found = "Entry point module was not found: fake_source_module.submodule"
        with pytest.raises(ValueError, match=module_not_found):
            _validate_entry_point(source_module_path, "fake_source_module.submodule")

        # catches AssertionError for module is not None
        submodule_not_found = "Entry point module was not found: source_module.fake_submodule"
        with pytest.raises(ValueError, match=submodule_not_found):
            _validate_entry_point(source_module_path, "source_module.fake_submodule")


@patch("tarfile.TarFile.add")
def test_tar_and_upload_to_code_location(mock_tar_add, aws_session):
    with tempfile.TemporaryDirectory() as temp_dir:
        source_module_path = Path(temp_dir, "source_module")
        source_module_path.mkdir()
        _tar_and_upload_to_code_location(source_module_path, aws_session, "code_location")
        mock_tar_add.assert_called_with(source_module_path, arcname="source_module")
        local, s3 = aws_session.upload_to_s3.call_args_list[0][0]
        assert local.endswith("source.tar.gz")
        assert s3 == "code_location/source.tar.gz"


@patch("braket.jobs.quantum_job_creation._process_local_source_module")
@patch("braket.jobs.quantum_job_creation._validate_entry_point")
@patch("braket.jobs.quantum_job_creation._validate_params")
def test_copy_checkpoints(
    mock_validate_input,
    mock_validate_entry_point,
    mock_process_local_source,
    aws_session,
    entry_point,
    device,
    checkpoint_config,
    generate_get_job_response,
):
    other_checkpoint_uri = "s3://amazon-braket-jobs/job-path/checkpoints"
    aws_session.get_job.return_value = generate_get_job_response(
        checkpointConfig={
            "s3Uri": other_checkpoint_uri,
        }
    )
    prepare_quantum_job(
        device=device,
        source_module="source_module",
        entry_point=entry_point,
        copy_checkpoints_from_job="other-job-arn",
        checkpoint_config=checkpoint_config,
        aws_session=aws_session,
    )
    aws_session.copy_s3_directory.assert_called_with(other_checkpoint_uri, checkpoint_config.s3Uri)


def test_invalid_input_parameters(entry_point, aws_session):
    error_message = (
        "'instance_config' should be of '<class 'braket.jobs.config.InstanceConfig'>' "
        "but user provided <class 'int'>."
    )
    with pytest.raises(ValueError, match=error_message):
        prepare_quantum_job(
            aws_session=aws_session,
            entry_point=entry_point,
            device="arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            source_module="alpha_test_job",
            hyperparameters={
                "param-1": "first parameter",
                "param-2": "second param",
            },
            instance_config=2,
        )


@pytest.mark.parametrize(
    "input_data, input_data_configs",
    (
        (
            "local/prefix",
            [
                {
                    "channelName": "input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://default-bucket-name/jobs/job-name/ts/data/input/prefix",
                        },
                    },
                }
            ],
        ),
        (
            "s3://my-bucket/my/prefix-",
            [
                {
                    "channelName": "input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://my-bucket/my/prefix-",
                        },
                    },
                }
            ],
        ),
        (
            S3DataSourceConfig(
                "s3://my-bucket/my/manifest.json",
                content_type="text/csv",
            ),
            [
                {
                    "channelName": "input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://my-bucket/my/manifest.json",
                        },
                    },
                    "contentType": "text/csv",
                }
            ],
        ),
        (
            {
                "local-input": "local/prefix",
                "s3-input": "s3://my-bucket/my/prefix-",
                "config-input": S3DataSourceConfig(
                    "s3://my-bucket/my/manifest.json",
                ),
            },
            [
                {
                    "channelName": "local-input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://default-bucket-name/jobs/job-name/ts/"
                            "data/local-input/prefix",
                        },
                    },
                },
                {
                    "channelName": "s3-input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://my-bucket/my/prefix-",
                        },
                    },
                },
                {
                    "channelName": "config-input",
                    "dataSource": {
                        "s3DataSource": {
                            "s3Uri": "s3://my-bucket/my/manifest.json",
                        },
                    },
                },
            ],
        ),
    ),
)
def test_process_input_data(aws_session, input_data, input_data_configs):
    job_name = "job-name"
    assert _process_input_data(input_data, job_name, aws_session, "ts") == input_data_configs



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.circuits import Gate, Instruction, Observable
from braket.circuits.noise_model import (
    CriteriaKey,
    CriteriaKeyResult,
    GateCriteria,
    ObservableCriteria,
)


@pytest.mark.parametrize(
    "gates, qubits",
    [
        (None, range(3)),
        (None, None),
        (Gate.H, range(3)),
        ([Gate.H], range(3)),
        ([Gate.I, Gate.H], range(3)),
        (Gate.H, 1),
        ([Gate.H], [1]),
        ([Gate.I, Gate.H], [1, 0]),
        (Gate.H, None),
    ],
)
def test_happy_case(gates, qubits):
    criteria = GateCriteria(gates=gates, qubits=qubits)
    assert criteria.applicable_key_types() == [CriteriaKey.QUBIT, CriteriaKey.GATE]
    if gates is None:
        assert CriteriaKeyResult.ALL == criteria.get_keys(CriteriaKey.GATE)
    else:
        assert Gate.H in criteria.get_keys(CriteriaKey.GATE)
    if qubits is None:
        assert CriteriaKeyResult.ALL == criteria.get_keys(CriteriaKey.QUBIT)
    else:
        assert 1 in criteria.get_keys(CriteriaKey.QUBIT)


@pytest.mark.parametrize(
    "gates, qubits",
    [
        (None, range(3)),
        (None, None),
        (Gate.H, range(3)),
        ([Gate.H], 1),
        ([Gate.H], [1]),
        ([Gate.H], [[1]]),
        ([Gate.CNot], [0, 1]),
        ([Gate.CNot], [[1, 2]]),
        ([Gate.CNot], [[3, 4], [5, 6]]),
        ([Gate.I, Gate.H], range(3)),
        ([Gate.I, Gate.H], None),
    ],
)
def test_serialization(gates, qubits):
    test_criteria = GateCriteria(gates=gates, qubits=qubits)
    serialized_criteria = test_criteria.to_dict()
    assert serialized_criteria["__class__"] == "GateCriteria"
    assert "gates" in serialized_criteria
    assert "qubits" in serialized_criteria
    deserialized_criteria = GateCriteria.from_dict(serialized_criteria)
    assert test_criteria == deserialized_criteria


@pytest.mark.parametrize(
    "gates, qubits, matching_instructions, non_matching_instructions",
    [
        (None, 1, [Instruction(Gate.H(), 1)], [Instruction(Gate.CNot(), [0, 1])]),
        (
            None,
            None,
            [Instruction(Gate.H(), 1), Instruction(Gate.CNot(), [0, 1])],
            [Instruction(Observable.X, 0)],
        ),
        (
            Gate.H,
            range(3),
            [Instruction(Gate.H(), 0), Instruction(Gate.H(), 1)],
            [
                Instruction(Gate.H(), 3),
                Instruction(Gate.I(), 1),
                [Instruction(Gate.H(), 0)],
            ],
        ),
        (Gate.H, 1, [Instruction(Gate.H(), 1)], [Instruction(Gate.H(), 0)]),
        ([Gate.H], [1], [Instruction(Gate.H(), 1)], [Instruction(Gate.H(), 0)]),
        (
            [Gate.CNot],
            [0, 1],
            [Instruction(Gate.CNot(), [0, 1])],
            [Instruction(Gate.CNot(), [1, 0])],
        ),
        (
            [Gate.CNot],
            [[0, 1], [1, 2]],
            [Instruction(Gate.CNot(), [0, 1]), Instruction(Gate.CNot(), [1, 2])],
            [Instruction(Gate.CNot(), [1, 0])],
        ),
        (
            [Gate.I, Gate.H],
            range(3),
            [Instruction(Gate.H(), 0), Instruction(Gate.I(), 2)],
            [Instruction(Gate.H(), 3), Instruction(Gate.X(), 1)],
        ),
        (
            Gate.H,
            None,
            [Instruction(Gate.H(), 0), Instruction(Gate.H(), 10)],
            [Instruction(Gate.I(), 1), Instruction(Gate.X(), 3)],
        ),
        (
            Gate.H,
            [],
            [Instruction(Gate.H(), 0), Instruction(Gate.H(), 10)],
            [Instruction(Gate.I(), 1), Instruction(Gate.X(), 3)],
        ),
    ],
)
def test_matcher(gates, qubits, matching_instructions, non_matching_instructions):
    criteria = GateCriteria(gates=gates, qubits=qubits)
    for instruction in matching_instructions:
        assert criteria.instruction_matches(instruction)
    for instruction in non_matching_instructions:
        assert not criteria.instruction_matches(instruction)


@pytest.mark.parametrize(
    "gates, qubits",
    [
        (Gate.CNot, [[0, 1], 2]),
    ],
)
@pytest.mark.xfail(raises=TypeError)
def test_invalid_type_params(gates, qubits):
    GateCriteria(gates=gates, qubits=qubits)


@pytest.mark.parametrize(
    "gates, qubits",
    [
        ([Gate.H, Gate.CNot], 1),
        (Gate.CNot, [[0, 1], [2]]),
    ],
)
@pytest.mark.xfail(raises=ValueError)
def test_invalid_value_params(gates, qubits):
    GateCriteria(gates=gates, qubits=qubits)


def test_representation():
    criteria = GateCriteria(gates=[Gate.I], qubits=0)
    str_representation = criteria.__repr__()
    assert str_representation == "GateCriteria(gates={'I'}, qubits={0})"


def test_string():
    criteria = GateCriteria(gates=[Gate.I], qubits=0)
    str_representation = criteria.__str__()
    assert str_representation == "GateCriteria({'I'}, {0})"


def test_get_keys_for_unknown_keytypes():
    criteria = GateCriteria(gates=[Gate.I], qubits=0)
    result = criteria.get_keys(CriteriaKey.UNITARY_GATE)
    assert len(result) == 0


@pytest.mark.parametrize(
    "criteria0, criteria1",
    [
        (GateCriteria(Gate.H, range(3)), GateCriteria(Gate.H, range(3))),
        (GateCriteria(Gate.H, [0, 1, 2]), GateCriteria(Gate.H, range(3))),
        (GateCriteria(Gate.H, [1, 2, 0]), GateCriteria(Gate.H, range(3))),
        (GateCriteria(Gate.H), GateCriteria(Gate.H, None)),
        (
            GateCriteria([Gate.H, Gate.I], range(3)),
            GateCriteria([Gate.I, Gate.H], range(3)),
        ),
    ],
)
def test_equal_criteria(criteria0, criteria1):
    assert criteria0 == criteria1


@pytest.mark.parametrize(
    "criteria0, criteria1",
    [
        (GateCriteria(Gate.H, range(3)), GateCriteria(Gate.I, range(3))),
        (GateCriteria(Gate.H, [1, 2, 3]), GateCriteria(Gate.H, range(3))),
        (GateCriteria(Gate.H), GateCriteria(Gate.H, range(3))),
        (GateCriteria(Gate.H, range(3)), float(3)),
        (GateCriteria(Gate.H, range(3)), ObservableCriteria(Observable.X, range(3))),
    ],
)
def test_not_equal_criteria(criteria0, criteria1):
    assert criteria0 != criteria1



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Any, ClassVar, Optional, Union

import numpy as np

from braket.circuits.free_parameter import FreeParameter
from braket.circuits.free_parameter_expression import FreeParameterExpression
from braket.circuits.parameterizable import Parameterizable
from braket.circuits.quantum_operator import QuantumOperator
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    SerializationProperties,
)
from braket.registers.qubit_set import QubitSet


class Noise(QuantumOperator):
    """Class `Noise` represents a noise channel that operates on one or multiple qubits. Noise
    are considered as building blocks of quantum circuits that simulate noise. It can be
    used as an operator in an `Instruction` object. It appears in the diagram when user prints
    a circuit with `Noise`. This class is considered the noise channel definition containing
    the metadata that defines what the noise channel is and what it does.
    """

    def __init__(self, qubit_count: Optional[int], ascii_symbols: Sequence[str]):
        """Initializes a `Noise` object.

        Args:
            qubit_count (Optional[int]): Number of qubits this noise channel interacts with.
            ascii_symbols (Sequence[str]): ASCII string symbols for this noise channel. These
                are used when printing a diagram of circuits. Length must be the same as
                `qubit_count`, and index ordering is expected to correlate with target ordering
                on the instruction.

        Raises:
            ValueError: `qubit_count` is less than 1, `ascii_symbols` are None, or
                length of `ascii_symbols` is not equal to `qubit_count`
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)

    @property
    def name(self) -> str:
        """Returns the name of the quantum operator

        Returns:
            str: The name of the quantum operator as a string
        """
        return self.__class__.__name__

    def to_ir(
        self,
        target: QubitSet,
        ir_type: IRType = IRType.JAQCD,
        serialization_properties: SerializationProperties | None = None,
    ) -> Any:
        """Returns IR object of quantum operator and target

        Args:
            target (QubitSet): target qubit(s)
            ir_type(IRType) : The IRType to use for converting the noise object to its
                IR representation. Defaults to IRType.JAQCD.
            serialization_properties (SerializationProperties | None): The serialization properties
                to use while serializing the object to the IR representation. The serialization
                properties supplied must correspond to the supplied `ir_type`. Defaults to None.

        Returns:
            Any: IR object of the quantum operator and target

        Raises:
            ValueError: If the supplied `ir_type` is not supported, or if the supplied serialization
            properties don't correspond to the `ir_type`.
        """
        if ir_type == IRType.JAQCD:
            return self._to_jaqcd(target)
        elif ir_type == IRType.OPENQASM:
            if serialization_properties and not isinstance(
                serialization_properties, OpenQASMSerializationProperties
            ):
                raise ValueError(
                    "serialization_properties must be of type OpenQASMSerializationProperties "
                    "for IRType.OPENQASM."
                )
            return self._to_openqasm(
                target, serialization_properties or OpenQASMSerializationProperties()
            )
        else:
            raise ValueError(f"Supplied ir_type {ir_type} is not supported.")

    def _to_jaqcd(self, target: QubitSet) -> Any:
        """Returns the JAQCD representation of the noise.

        Args:
            target (QubitSet): target qubit(s).

        Returns:
            Any: JAQCD object representing the noise.
        """
        raise NotImplementedError("to_jaqcd has not been implemented yet.")

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        """Returns the openqasm string representation of the noise.

        Args:
            target (QubitSet): target qubit(s).
            serialization_properties (OpenQASMSerializationProperties): The serialization properties
                to use while serializing the object to the IR representation.

        Returns:
            str: Representing the openqasm representation of the noise.
        """
        raise NotImplementedError("to_openqasm has not been implemented yet.")

    def to_matrix(self, *args, **kwargs) -> Iterable[np.ndarray]:
        """Returns a list of matrices defining the Kraus matrices of the noise channel.

        Returns:
            Iterable[ndarray]: list of matrices defining the Kraus matrices of the noise channel.
        """
        raise NotImplementedError("to_matrix has not been implemented yet.")

    def __eq__(self, other: Noise):
        if isinstance(other, Noise):
            return self.name == other.name
        return False

    def __repr__(self):
        return f"{self.name}('qubit_count': {self.qubit_count})"

    @classmethod
    def from_dict(cls, noise: dict) -> Noise:
        """Converts a dictionary representing an object of this class into an instance of
        this class.

        Args:
            noise (dict): A dictionary representation of an object of this class.

        Returns:
            Noise: An object of this class that corresponds to the passed in dictionary.
        """
        if "__class__" in noise:
            noise_name = noise["__class__"]
            noise_cls = getattr(cls, noise_name)
            return noise_cls.from_dict(noise)
        raise NotImplementedError

    @classmethod
    def register_noise(cls, noise: type[Noise]) -> None:
        """Register a noise implementation by adding it into the Noise class.

        Args:
            noise (type[Noise]): Noise class to register.
        """
        setattr(cls, noise.__name__, noise)


class SingleProbabilisticNoise(Noise, Parameterizable):
    """Class `SingleProbabilisticNoise` represents the bit/phase flip noise channel on N qubits
    parameterized by a single probability.
    """

    def __init__(
        self,
        probability: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
        max_probability: float = 0.5,
    ):
        """Initializes a `SingleProbabilisticNoise`.

        Args:
            probability (Union[FreeParameterExpression, float]): The probability that the
                noise occurs.
            qubit_count (Optional[int]): The number of qubits to apply noise.
            ascii_symbols (Sequence[str]): ASCII string symbols for the noise. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.
            max_probability (float): Maximum allowed probability of the noise channel. Default: 0.5

        Raises:
            ValueError: If the `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`, `probability` is not `float` or
                `FreeParameterExpression`, `probability` > 1/2, or `probability` < 0
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)

        if not isinstance(probability, FreeParameterExpression):
            _validate_param_value(probability, "probability", max_probability)
        self._probability = probability

    @property
    def probability(self) -> float:
        """The probability that parametrizes the noise channel.

        Returns:
            float: The probability that parametrizes the noise channel.
        """
        return self._probability

    def __repr__(self):
        return f"{self.name}('probability': {self.probability}, 'qubit_count': {self.qubit_count})"

    def __str__(self):
        return f"{self.name}({self.probability})"

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameter expressions
            or fixed values associated with the object.
        """
        return [self._probability]

    def __eq__(self, other: SingleProbabilisticNoise):
        if isinstance(other, SingleProbabilisticNoise):
            return self.name == other.name and self.probability == other.probability
        return False

    def bind_values(self, **kwargs) -> SingleProbabilisticNoise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            SingleProbabilisticNoise: A new Noise object of the same type with the requested
            parameters bound.

        Raises:
            NotImplementedError: Subclasses should implement this function.
        """
        raise NotImplementedError

    def to_dict(self) -> dict:
        """Converts this object into a dictionary representation.

        Returns:
            dict: A dictionary object that represents this object. It can be converted back
            into this object using the `from_dict()` method.
        """
        return {
            "__class__": self.__class__.__name__,
            "probability": _parameter_to_dict(self.probability),
            "qubit_count": self.qubit_count,
            "ascii_symbols": self.ascii_symbols,
        }


class SingleProbabilisticNoise_34(SingleProbabilisticNoise):
    """Class `SingleProbabilisticNoise` represents the Depolarizing and TwoQubitDephasing noise
    channels parameterized by a single probability.
    """

    def __init__(
        self,
        probability: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Initializes a `SingleProbabilisticNoise_34`.

        Args:
            probability (Union[FreeParameterExpression, float]): The probability that the
                noise occurs.
            qubit_count (Optional[int]): The number of qubits to apply noise.
            ascii_symbols (Sequence[str]): ASCII string symbols for the noise. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.

        Raises:
            ValueError: If the `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`, `probability` is not `float` or
                `FreeParameterExpression`, `probability` > 3/4, or `probability` < 0
        """
        super().__init__(
            probability=probability,
            qubit_count=qubit_count,
            ascii_symbols=ascii_symbols,
            max_probability=0.75,
        )


class SingleProbabilisticNoise_1516(SingleProbabilisticNoise):
    """Class `SingleProbabilisticNoise` represents the TwoQubitDepolarizing noise channel
    parameterized by a single probability.
    """

    def __init__(
        self,
        probability: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Initializes a `SingleProbabilisticNoise_1516`.

        Args:
            probability (Union[FreeParameterExpression, float]): The probability that the
                noise occurs.
            qubit_count (Optional[int]): The number of qubits to apply noise.
            ascii_symbols (Sequence[str]): ASCII string symbols for the noise. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.

        Raises:
            ValueError: If the `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`, `probability` is not `float` or
                `FreeParameterExpression`, `probability` > 15/16, or `probability` < 0
        """
        super().__init__(
            probability=probability,
            qubit_count=qubit_count,
            ascii_symbols=ascii_symbols,
            max_probability=0.9375,
        )


class MultiQubitPauliNoise(Noise, Parameterizable):
    """Class `MultiQubitPauliNoise` represents a general multi-qubit Pauli channel,
    parameterized by up to 4**N - 1 probabilities.
    """

    _allowed_substrings: ClassVar = {"I", "X", "Y", "Z"}

    def __init__(
        self,
        probabilities: dict[str, Union[FreeParameterExpression, float]],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """[summary]

        Args:
            probabilities (dict[str, Union[FreeParameterExpression, float]]): A dictionary with
                Pauli strings as keys and the probabilities as values, i.e. {"XX": 0.1. "IZ": 0.2}.
            qubit_count (Optional[int]): The number of qubits the Pauli noise acts on.
            ascii_symbols (Sequence[str]): ASCII string symbols for the noise. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.

        Raises:
            ValueError: If
                `qubit_count` < 1,
                `ascii_symbols` is `None`,
                `ascii_symbols` length != `qubit_count`,
                `probabilities` are not `float`s or FreeParameterExpressions,
                any of `probabilities` > 1 or `probabilities` < 0,
                the sum of all probabilities is > 1,
                if "II" is specified as a Pauli string,
                if any Pauli string contains invalid strings,
                or if the length of probabilities is greater than 4**qubit_count.
            TypeError: If the type of the dictionary keys are not strings.
                If the probabilities are not floats.
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)
        self._probabilities = probabilities

        if not probabilities:
            raise ValueError("Pauli dictionary must not be empty.")

        identity = self.qubit_count * "I"
        if identity in probabilities:
            raise ValueError(
                f"{identity} is not allowed as a key. Please enter only non-identity Pauli strings."
            )

        total_prob = 0
        for pauli_string, prob in probabilities.items():
            MultiQubitPauliNoise._validate_pauli_string(
                pauli_string, self.qubit_count, self._allowed_substrings
            )
            if not isinstance(prob, FreeParameterExpression):
                _validate_param_value(prob, f"probability for {pauli_string}")
                total_prob += prob
        if not (1.0 >= total_prob >= 0.0):
            raise ValueError(
                "Total probability must be a real number in the interval [0, 1]. "
                f"Total probability was {total_prob}."
            )

    @classmethod
    def _validate_pauli_string(
        cls, pauli_str: str, qubit_count: int, allowed_substrings: set[str]
    ) -> None:
        if not isinstance(pauli_str, str):
            raise TypeError(f"Type of {pauli_str} was not a string.")
        if len(pauli_str) != qubit_count:
            raise ValueError(
                "Length of each Pauli string must be equal to number of qubits. "
                f"{pauli_str} had length {len(pauli_str)} instead of length {qubit_count}."
            )
        if not set(pauli_str) <= allowed_substrings:
            raise ValueError(
                "Strings must be Pauli strings consisting of only [I, X, Y, Z]. "
                f"Received {pauli_str}."
            )

    def __repr__(self):
        return (
            f"{self.name}('probabilities' : {self._probabilities}, "
            f"'qubit_count': {self.qubit_count})"
        )

    def __str__(self):
        return f"{self.name}({self._probabilities})"

    def __eq__(self, other: MultiQubitPauliNoise):
        if isinstance(other, MultiQubitPauliNoise):
            return self.name == other.name and self._probabilities == other._probabilities
        return False

    @property
    def probabilities(self) -> dict[str, float]:
        """A map of a Pauli string to its corresponding probability.

        Returns:
            dict[str, float]: A map of a Pauli string to its corresponding probability.
        """
        return self._probabilities

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.

        Parameters are in alphabetical order of the Pauli strings in `probabilities`.

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameter expressions
            or fixed values associated with the object.
        """
        return [
            self._probabilities[pauli_string] for pauli_string in sorted(self._probabilities.keys())
        ]

    def bind_values(self, **kwargs) -> MultiQubitPauliNoise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            MultiQubitPauliNoise: A new Noise object of the same type with the requested
            parameters bound.

        Raises:
            NotImplementedError: Subclasses should implement this function.
        """
        raise NotImplementedError

    def to_dict(self) -> dict:
        """Converts this object into a dictionary representation.

        Returns:
            dict: A dictionary object that represents this object. It can be converted back
            into this object using the `from_dict()` method.
        """
        probabilities = {}
        for pauli_string, prob in self._probabilities.items():
            probabilities[pauli_string] = _parameter_to_dict(prob)
        return {
            "__class__": self.__class__.__name__,
            "probabilities": probabilities,
            "qubit_count": self.qubit_count,
            "ascii_symbols": self.ascii_symbols,
        }


class PauliNoise(Noise, Parameterizable):
    """Class `PauliNoise` represents the a single-qubit Pauli noise channel
    acting on one qubit. It is parameterized by three probabilities.
    """

    def __init__(
        self,
        probX: Union[FreeParameterExpression, float],
        probY: Union[FreeParameterExpression, float],
        probZ: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Initializes a `PauliNoise`.

        Args:
            probX (Union[FreeParameterExpression, float]): The X coefficient of the Kraus operators
                in the channel.
            probY (Union[FreeParameterExpression, float]): The Y coefficient of the Kraus operators
                in the channel.
            probZ (Union[FreeParameterExpression, float]): The Z coefficient of the Kraus operators
                in the channel.
            qubit_count (Optional[int]): The number of qubits to apply noise.
            ascii_symbols (Sequence[str]): ASCII string symbols for the noise. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.

        Raises:
            ValueError: If the `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`, `probX` or `probY` or `probZ`
                is not `float` or FreeParameterExpression, `probX` or `probY` or `probZ` > 1.0, or
                `probX` or `probY` or `probZ` < 0.0, or `probX`+`probY`+`probZ` > 1
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)

        total = 0
        total += PauliNoise._get_param_float(probX, "probX")
        total += PauliNoise._get_param_float(probY, "probY")
        total += PauliNoise._get_param_float(probZ, "probZ")
        if total > 1:
            raise ValueError("the sum of probX, probY, probZ cannot be larger than 1")
        self._parameters = [probX, probY, probZ]

    @staticmethod
    def _get_param_float(param: Union[FreeParameterExpression, float], param_name: str) -> float:
        """Validates the value of a probability and returns its value.

        If param is a free parameter expression, this method returns 0.

        Args:
            param (Union[FreeParameterExpression, float]): The probability to validate
            param_name (str): The name of the probability parameter

        Returns:
            float: The value of the parameter, or 0 if it is a free parameter expression
        """
        if isinstance(param, FreeParameterExpression):
            return 0
        else:
            _validate_param_value(param, param_name)
            return float(param)

    @property
    def probX(self) -> Union[FreeParameterExpression, float]:
        """The probability of a Pauli X error.

        Returns:
            Union[FreeParameterExpression, float]: The probability of a Pauli X error.
        """
        return self._parameters[0]

    @property
    def probY(self) -> Union[FreeParameterExpression, float]:
        """The probability of a Pauli Y error.

        Returns:
            Union[FreeParameterExpression, float]: The probability of a Pauli Y error.
        """
        return self._parameters[1]

    @property
    def probZ(self) -> Union[FreeParameterExpression, float]:
        """The probability of a Pauli Z error.

        Returns:
            Union[FreeParameterExpression, float]: The probability of a Pauli Z error.
        """
        return self._parameters[2]

    def __repr__(self):
        return (
            f"{self.name}("
            f"'probX': {self._parameters[0]}, "
            f"'probY': {self._parameters[1]}, "
            f"'probZ': {self._parameters[2]}, "
            f"'qubit_count': {self.qubit_count}"
            f")"
        )

    def __str__(self):
        return f"{self.name}({self._parameters[0]}, {self._parameters[1]}, {self._parameters[2]})"

    def __eq__(self, other: PauliNoise):
        if isinstance(other, PauliNoise):
            return self.name == other.name and self._parameters == other._parameters
        return False

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.

        Parameters are in the order [probX, probY, probZ]

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameter expressions
            or fixed values associated with the object.
        """
        return self._parameters

    def bind_values(self, **kwargs) -> PauliNoise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            PauliNoise: A new Noise object of the same type with the requested
            parameters bound.

        Raises:
            NotImplementedError: Subclasses should implement this function.
        """
        raise NotImplementedError

    def to_dict(self) -> dict:
        """Converts this object into a dictionary representation.

        Returns:
            dict: A dictionary object that represents this object. It can be converted back
            into this object using the `from_dict()` method.
        """
        return {
            "__class__": self.__class__.__name__,
            "probX": _parameter_to_dict(self.probX),
            "probY": _parameter_to_dict(self.probY),
            "probZ": _parameter_to_dict(self.probZ),
            "qubit_count": self.qubit_count,
            "ascii_symbols": self.ascii_symbols,
        }


class DampingNoise(Noise, Parameterizable):
    """Class `DampingNoise` represents a damping noise channel
    on N qubits parameterized by gamma.
    """

    def __init__(
        self,
        gamma: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Initalizes a `DampingNoise`.

        Args:
            gamma (Union[FreeParameterExpression, float]): Probability of damping.
            qubit_count (Optional[int]): The number of qubits to apply noise.
            ascii_symbols (Sequence[str]): ASCII string symbols for the noise. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.

        Raises:
                ValueError: If
                    `qubit_count` < 1,
                    `ascii_symbols` is `None`,
                    `len(ascii_symbols)` != `qubit_count`,
                    `gamma` is not `float` or `FreeParameterExpression`,
                    or `gamma` > 1.0 or `gamma` < 0.0.
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)

        if not isinstance(gamma, FreeParameterExpression):
            _validate_param_value(gamma, "gamma")
        self._gamma = gamma

    @property
    def gamma(self) -> float:
        """Probability of damping.

        Returns:
            float: Probability of damping.
        """
        return self._gamma

    def __repr__(self):
        return f"{self.name}('gamma': {self.gamma}, 'qubit_count': {self.qubit_count})"

    def __str__(self):
        return f"{self.name}({self.gamma})"

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameter expressions
            or fixed values associated with the object.
        """
        return [self._gamma]

    def __eq__(self, other: DampingNoise):
        if isinstance(other, DampingNoise):
            return self.name == other.name and self.gamma == other.gamma
        return False

    def bind_values(self, **kwargs) -> DampingNoise:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            DampingNoise: A new Noise object of the same type with the requested
            parameters bound.

        Raises:
            NotImplementedError: Subclasses should implement this function.
        """
        raise NotImplementedError

    def to_dict(self) -> dict:
        """Converts this object into a dictionary representation.

        Returns:
            dict: A dictionary object that represents this object. It can be converted back
            into this object using the `from_dict()` method.
        """
        return {
            "__class__": self.__class__.__name__,
            "gamma": _parameter_to_dict(self.gamma),
            "qubit_count": self.qubit_count,
            "ascii_symbols": self.ascii_symbols,
        }


class GeneralizedAmplitudeDampingNoise(DampingNoise):
    """Class `GeneralizedAmplitudeDampingNoise` represents the generalized amplitude damping
    noise channel on N qubits parameterized by gamma and probability.
    """

    def __init__(
        self,
        gamma: Union[FreeParameterExpression, float],
        probability: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Inits a `GeneralizedAmplitudeDampingNoise`.

        Args:
            gamma (Union[FreeParameterExpression, float]): Probability of damping.
            probability (Union[FreeParameterExpression, float]): Probability of the system being
                excited by the environment.
            qubit_count (Optional[int]): The number of qubits to apply noise.
            ascii_symbols (Sequence[str]): ASCII string symbols for the noise. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.

        Raises:
                ValueError: If
                    `qubit_count` < 1,
                    `ascii_symbols` is `None`,
                    `len(ascii_symbols)` != `qubit_count`,
                    `probability` or `gamma` is not `float` or `FreeParameterExpression`,
                    `probability` > 1.0 or `probability` < 0.0,
                    or `gamma` > 1.0 or `gamma` < 0.0.
        """
        super().__init__(gamma=gamma, qubit_count=qubit_count, ascii_symbols=ascii_symbols)

        if not isinstance(probability, FreeParameterExpression):
            _validate_param_value(probability, "probability")
        self._probability = probability

    @property
    def probability(self) -> float:
        """Probability of the system being excited by the environment.

        Returns:
            float: Probability of the system being excited by the environment.
        """
        return self._probability

    def __repr__(self):
        return (
            f"{self.name}('gamma': {self.gamma}, "
            f"'probability': {self.probability}, "
            f"'qubit_count': {self.qubit_count})"
        )

    def __str__(self):
        return f"{self.name}({self.gamma}, {self.probability})"

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameter
        expressions or bound values.

        Parameters are in the order [gamma, probability]

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameter expressions
            or fixed values associated with the object.
        """
        return [self.gamma, self.probability]

    def __eq__(self, other: GeneralizedAmplitudeDampingNoise):
        if isinstance(other, GeneralizedAmplitudeDampingNoise):
            return (
                self.name == other.name
                and self.gamma == other.gamma
                and self.probability == other.probability
            )
        return False

    def to_dict(self) -> dict:
        """Converts this object into a dictionary representation.

        Returns:
            dict: A dictionary object that represents this object. It can be converted back
            into this object using the `from_dict()` method.
        """
        return {
            "__class__": self.__class__.__name__,
            "gamma": _parameter_to_dict(self.gamma),
            "probability": _parameter_to_dict(self.probability),
            "qubit_count": self.qubit_count,
            "ascii_symbols": self.ascii_symbols,
        }


def _validate_param_value(
    parameter: Union[FreeParameterExpression, float], param_name: str, maximum: float = 1.0
) -> None:
    """Validates the value of a given parameter.

    Args:
        parameter (Union[FreeParameterExpression, float]): The parameter to validate
        param_name (str): The name of the parameter
        maximum (float): The maximum value of the parameter. Default: 1.0
    """
    if not isinstance(parameter, float):
        raise TypeError(f"{param_name} must be float type")
    if not (maximum >= parameter >= 0.0):
        raise ValueError(f"{param_name} must be a real number in the interval [0, {maximum}]")


def _parameter_to_dict(parameter: Union[FreeParameter, float]) -> Union[dict, float]:
    """Converts a parameter to a dictionary if it's a FreeParameter, otherwise returns the float.

    Args:
        parameter(Union[FreeParameter, float]): The parameter to convert.

    Returns:
        Union[dict, float]: A dictionary representation of a FreeParameter if the parameter
        is a FreeParameter, otherwise returns the float.
    """
    if isinstance(parameter, FreeParameter):
        return parameter.to_dict()
    return parameter



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Optional

import numpy as np

from braket.circuits.operator import Operator


class QuantumOperator(Operator):
    """A quantum operator is the definition of a quantum operation for a quantum device."""

    def __init__(self, qubit_count: Optional[int], ascii_symbols: Sequence[str]):
        """Initializes a `QuantumOperator`.

        Args:
            qubit_count (Optional[int]): Number of qubits this quantum operator acts on.
                If all instances of the operator act on the same number of qubits, this argument
                should be ``None``, and ``fixed_qubit_count`` should be implemented to return
                the qubit count; if ``fixed_qubit_count`` is implemented and an int is passed in,
                it must equal ``fixed_qubit_count``, or instantiation will raise a ValueError.
                An int must be passed in if instances can have a varying number of qubits,
                in which case ``fixed_qubit_count`` should not be implemented,
            ascii_symbols (Sequence[str]): ASCII string symbols for the quantum operator.
                These are used when printing a diagram of circuits.
                Length must be the same as `qubit_count`, and index ordering is expected
                to correlate with target ordering on the instruction.
                For instance, if CNOT instruction has the control qubit on the first index and
                target qubit on the second index. Then ASCII symbols would have ["C", "X"] to
                correlate a symbol with that index.

        Raises:
            TypeError: `qubit_count` is not an int
            ValueError: `qubit_count` is less than 1, `ascii_symbols` are `None`,
                ``fixed_qubit_count`` is implemented and and not equal to ``qubit_count``,
                or ``len(ascii_symbols) != qubit_count``
        """
        fixed_qubit_count = self.fixed_qubit_count()
        if fixed_qubit_count is NotImplemented:
            self._qubit_count = qubit_count
        else:
            if qubit_count and qubit_count != fixed_qubit_count:
                raise ValueError(
                    f"Provided qubit count {qubit_count}"
                    "does not equal fixed qubit count {fixed_qubit_count}"
                )
            self._qubit_count = fixed_qubit_count

        if not isinstance(self._qubit_count, int):
            raise TypeError(f"qubit_count, {self._qubit_count}, must be an integer")

        if self._qubit_count < 1:
            raise ValueError(f"qubit_count, {self._qubit_count}, must be greater than zero")

        if ascii_symbols is None:
            raise ValueError("ascii_symbols must not be None")

        if len(ascii_symbols) != self._qubit_count:
            msg = (
                f"ascii_symbols, {ascii_symbols},"
                f" length must equal qubit_count, {self._qubit_count}"
            )
            raise ValueError(msg)
        self._ascii_symbols = tuple(ascii_symbols)

    @staticmethod
    def fixed_qubit_count() -> int:
        """Returns the number of qubits this quantum operator acts on,
        if instances are guaranteed to act on the same number of qubits.

        If different instances can act on a different number of qubits,
        this method returns ``NotImplemented``.

        Returns:
            int: The number of qubits this quantum operator acts on.
        """
        return NotImplemented

    @property
    def qubit_count(self) -> int:
        """int: The number of qubits this quantum operator acts on."""
        return self._qubit_count

    @property
    def ascii_symbols(self) -> tuple[str, ...]:
        """tuple[str, ...]: Returns the ascii symbols for the quantum operator."""
        return self._ascii_symbols

    @property
    def name(self) -> str:
        """Returns the name of the quantum operator

        Returns:
            str: The name of the quantum operator as a string
        """
        return self.__class__.__name__

    def to_ir(self, *args: Any, **kwargs: Any) -> Any:
        """Returns IR representation of quantum operator.

        Args:
            *args (Any): Not Implemented.
            **kwargs (Any): Not Implemented.

        Raises:
            NotImplementError: Not Implemented.

        Returns:
            Any: The the canonical intermediate representation of the operator.
        """
        raise NotImplementedError("to_ir has not been implemented yet.")

    def to_matrix(self, *args: Any, **kwargs: Any) -> np.ndarray:
        """Returns a matrix representation of the quantum operator.

        Args:
            *args (Any): Not Implemented.
            **kwargs (Any): Not Implemented.

        Raises:
            NotImplementError: Not Implemented.

        Returns:
            np.ndarray: A matrix representation of the quantum operator
        """
        raise NotImplementedError("to_matrix has not been implemented yet.")

    def matrix_equivalence(self, other: QuantumOperator) -> bool:
        """Whether the matrix form of two quantum operators are equivalent

        Args:
            other (QuantumOperator): Quantum operator instance to compare this quantum operator to

        Returns:
            bool: If matrix forms of this quantum operator and the other quantum operator
            are equivalent
        """
        if not isinstance(other, QuantumOperator):
            return False
        try:
            return np.allclose(self.to_matrix(), other.to_matrix())
        except ValueError:
            return False

    def __repr__(self):
        return f"{self.name}('qubit_count': {self.qubit_count})"



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket.circuits.noise_model.circuit_instruction_criteria import (  # noqa: F401
    CircuitInstructionCriteria,
)
from braket.circuits.noise_model.criteria import (  # noqa: F401
    Criteria,
    CriteriaKey,
    CriteriaKeyResult,
)
from braket.circuits.noise_model.gate_criteria import GateCriteria  # noqa: F401
from braket.circuits.noise_model.initialization_criteria import InitializationCriteria  # noqa: F401
from braket.circuits.noise_model.noise_model import NoiseModel, NoiseModelInstruction  # noqa: F401
from braket.circuits.noise_model.observable_criteria import ObservableCriteria  # noqa: F401
from braket.circuits.noise_model.qubit_initialization_criteria import (  # noqa: F401
    QubitInitializationCriteria,
)
from braket.circuits.noise_model.result_type_criteria import ResultTypeCriteria  # noqa: F401
from braket.circuits.noise_model.unitary_gate_criteria import UnitaryGateCriteria  # noqa: F401



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Any

from braket.circuits.quantum_operator import QuantumOperator
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    SerializationProperties,
)
from braket.registers.qubit_set import QubitSet


class Measure(QuantumOperator):
    """Class `Measure` represents a measure operation on targeted qubits"""

    def __init__(self, **kwargs):
        """Inits a `Measure`.

        Raises:
            ValueError: `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`
        """
        super().__init__(qubit_count=1, ascii_symbols=["M"])
        self._target_index = kwargs.get("index")

    @property
    def ascii_symbols(self) -> tuple[str]:
        """tuple[str]: Returns the ascii symbols for the measure."""
        return self._ascii_symbols

    def to_ir(
        self,
        target: QubitSet | None = None,
        ir_type: IRType = IRType.OPENQASM,
        serialization_properties: SerializationProperties | None = None,
        **kwargs,
    ) -> Any:
        """Returns IR object of the measure operator.

        Args:
            target (QubitSet | None): target qubit(s). Defaults to None
            ir_type(IRType) : The IRType to use for converting the measure object to its
                IR representation. Defaults to IRType.OpenQASM.
            serialization_properties (SerializationProperties | None): The serialization properties
                to use while serializing the object to the IR representation. The serialization
                properties supplied must correspond to the supplied `ir_type`. Defaults to None.

        Returns:
            Any: IR object of the measure operator.

        Raises:
            ValueError: If the supplied `ir_type` is not supported.
        """
        if ir_type == IRType.JAQCD:
            return self._to_jaqcd()
        elif ir_type == IRType.OPENQASM:
            return self._to_openqasm(
                target, serialization_properties or OpenQASMSerializationProperties() ** kwargs
            )
        else:
            raise ValueError(f"supplied ir_type {ir_type} is not supported.")

    def _to_jaqcd(self) -> Any:
        """Returns the JAQCD representation of the measure."""
        raise NotImplementedError("measure instructions are not supported with JAQCD.")

    def _to_openqasm(
        self, target: QubitSet, serialization_properties: OpenQASMSerializationProperties
    ) -> str:
        """Returns the openqasm string representation of the measure."""
        target_qubits = [serialization_properties.format_target(int(qubit)) for qubit in target]
        instructions = []
        for idx, qubit in enumerate(target_qubits):
            bit_index = (
                self._target_index if self._target_index and len(target_qubits) == 1 else idx
            )
            instructions.append(f"b[{bit_index}] = measure {qubit};")

        return "\n".join(instructions)

    def __eq__(self, other: Measure):
        return isinstance(other, Measure)

    def __repr__(self):
        return self.name



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import textwrap
import warnings
from typing import Any, Dict, Optional
from unittest.mock import Mock, patch

import pytest
from pydantic.v1 import create_model  # This is temporary for defining properties below

import braket.ir as ir
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.atom_arrangement import AtomArrangement
from braket.ahs.hamiltonian import Hamiltonian
from braket.annealing import Problem, ProblemType
from braket.circuits import Circuit, FreeParameter, Gate, Noise
from braket.circuits.noise_model import GateCriteria, NoiseModel, NoiseModelInstruction
from braket.device_schema import DeviceActionType, DeviceCapabilities
from braket.device_schema.openqasm_device_action_properties import OpenQASMDeviceActionProperties
from braket.devices import LocalSimulator, local_simulator
from braket.ir.openqasm import Program
from braket.simulator import BraketSimulator
from braket.task_result import AnnealingTaskResult, GateModelTaskResult
from braket.task_result.analog_hamiltonian_simulation_task_result_v1 import (
    AnalogHamiltonianSimulationTaskResult,
)
from braket.tasks import AnnealingQuantumTaskResult, GateModelQuantumTaskResult
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (
    AnalogHamiltonianSimulationQuantumTaskResult,
)

GATE_MODEL_RESULT = GateModelTaskResult(
    **{
        "measurements": [[0, 0], [0, 0], [0, 0], [1, 1]],
        "measuredQubits": [0, 1],
        "taskMetadata": {
            "braketSchemaHeader": {"name": "braket.task_result.task_metadata", "version": "1"},
            "id": "task_arn",
            "shots": 100,
            "deviceId": "default",
        },
        "additionalMetadata": {
            "action": {
                "braketSchemaHeader": {"name": "braket.ir.jaqcd.program", "version": "1"},
                "instructions": [{"control": 0, "target": 1, "type": "cnot"}],
            },
        },
    }
)

ANNEALING_RESULT = AnnealingTaskResult(
    **{
        "solutions": [[-1, -1, -1, -1], [1, -1, 1, 1], [1, -1, -1, 1]],
        "solutionCounts": [3, 2, 4],
        "values": [0.0, 1.0, 2.0],
        "variableCount": 4,
        "taskMetadata": {
            "id": "task_arn",
            "shots": 100,
            "deviceId": "device_id",
        },
        "additionalMetadata": {
            "action": {
                "type": "ISING",
                "linear": {"0": 0.3333, "1": -0.333, "4": -0.333, "5": 0.333},
                "quadratic": {"0,4": 0.667, "0,5": -1.0, "1,4": 0.667, "1,5": 0.667},
            },
            "dwaveMetadata": {
                "activeVariables": [0],
                "timing": {
                    "qpuSamplingTime": 100,
                    "qpuAnnealTimePerSample": 20,
                    "qpuAccessTime": 10917,
                    "qpuAccessOverheadTime": 3382,
                    "qpuReadoutTimePerSample": 274,
                    "qpuProgrammingTime": 9342,
                    "qpuDelayTimePerSample": 21,
                    "postProcessingOverheadTime": 117,
                    "totalPostProcessingTime": 117,
                    "totalRealTime": 10917,
                    "runTimeChip": 1575,
                    "annealTimePerRun": 20,
                    "readoutTimePerRun": 274,
                },
            },
        },
    }
)

AHS_RESULT = AnalogHamiltonianSimulationTaskResult(
    **{
        "taskMetadata": {
            "id": "rydberg",
            "shots": 100,
            "deviceId": "rydbergLocalSimulator",
        },
    }
)


class DummyCircuitSimulator(BraketSimulator):
    def run(
        self,
        program: ir.jaqcd.Program,
        qubits: int,
        shots: Optional[int],
        inputs: Optional[Dict[str, float]],
        *args,
        **kwargs,
    ) -> Dict[str, Any]:
        self._shots = shots
        self._qubits = qubits
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        return DeviceCapabilities.parse_obj(
            {
                "service": {
                    "executionWindows": [
                        {
                            "executionDay": "Everyday",
                            "windowStartHour": "11:00",
                            "windowEndHour": "12:00",
                        }
                    ],
                    "shotsRange": [1, 10],
                },
                "action": {
                    "braket.ir.openqasm.program": {
                        "actionType": "braket.ir.openqasm.program",
                        "version": ["1"],
                    },
                    "braket.ir.jaqcd.program": {
                        "actionType": "braket.ir.jaqcd.program",
                        "version": ["1"],
                    },
                },
                "deviceParameters": {},
            }
        )


class DummyJaqcdSimulator(BraketSimulator):
    def run(
        self, program: ir.jaqcd.Program, qubits: int, shots: Optional[int], *args, **kwargs
    ) -> Dict[str, Any]:
        if not isinstance(program, ir.jaqcd.Program):
            raise TypeError("Not a Jaqcd program")
        self._shots = shots
        self._qubits = qubits
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        return DeviceCapabilities.parse_obj(
            {
                "service": {
                    "executionWindows": [
                        {
                            "executionDay": "Everyday",
                            "windowStartHour": "11:00",
                            "windowEndHour": "12:00",
                        }
                    ],
                    "shotsRange": [1, 10],
                },
                "action": {
                    "braket.ir.jaqcd.program": {
                        "actionType": "braket.ir.jaqcd.program",
                        "version": ["1"],
                    },
                },
                "deviceParameters": {},
            }
        )

    def assert_shots(self, shots):
        assert self._shots == shots

    def assert_qubits(self, qubits):
        assert self._qubits == qubits


class DummyProgramSimulator(BraketSimulator):
    def run(
        self,
        openqasm_ir: Program,
        shots: int = 0,
        batch_size: int = 1,
    ) -> GateModelTaskResult:
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        device_properties = DeviceCapabilities.parse_obj(
            {
                "service": {
                    "executionWindows": [
                        {
                            "executionDay": "Everyday",
                            "windowStartHour": "00:00",
                            "windowEndHour": "23:59:59",
                        }
                    ],
                    "shotsRange": [1, 10],
                },
                "action": {
                    "braket.ir.openqasm.program": {
                        "actionType": "braket.ir.openqasm.program",
                        "version": ["1"],
                    }
                },
                "deviceParameters": {},
            }
        )
        oq3_action = OpenQASMDeviceActionProperties.parse_raw(
            json.dumps(
                {
                    "actionType": "braket.ir.openqasm.program",
                    "version": ["1"],
                    "supportedOperations": ["rx", "ry", "h", "cy", "cnot", "unitary"],
                    "supportedResultTypes": [
                        {"name": "StateVector", "observables": None, "minShots": 0, "maxShots": 0},
                    ],
                    "supportedPragmas": [
                        "braket_unitary_matrix",
                        "braket_result_type_sample",
                        "braket_result_type_expectation",
                        "braket_result_type_variance",
                        "braket_result_type_probability",
                        "braket_result_type_state_vector",
                    ],
                }
            )
        )
        device_properties.action[DeviceActionType.OPENQASM] = oq3_action
        return device_properties


class DummyProgramDensityMatrixSimulator(BraketSimulator):
    def run(
        self, program: ir.openqasm.Program, shots: Optional[int], *args, **kwargs
    ) -> Dict[str, Any]:
        self._shots = shots
        return GATE_MODEL_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        device_properties = DeviceCapabilities.parse_obj(
            {
                "service": {
                    "executionWindows": [
                        {
                            "executionDay": "Everyday",
                            "windowStartHour": "11:00",
                            "windowEndHour": "12:00",
                        }
                    ],
                    "shotsRange": [1, 10],
                },
                "action": {},
                "deviceParameters": {},
            }
        )
        oq3_action = OpenQASMDeviceActionProperties.parse_raw(
            json.dumps(
                {
                    "actionType": "braket.ir.openqasm.program",
                    "version": ["1"],
                    "supportedOperations": ["rx", "ry", "h", "cy", "cnot", "unitary"],
                    "supportedResultTypes": [
                        {"name": "StateVector", "observables": None, "minShots": 0, "maxShots": 0},
                    ],
                    "supportedPragmas": [
                        "braket_noise_bit_flip",
                        "braket_noise_depolarizing",
                        "braket_noise_kraus",
                        "braket_noise_pauli_channel",
                        "braket_noise_generalized_amplitude_damping",
                        "braket_noise_amplitude_damping",
                        "braket_noise_phase_flip",
                        "braket_noise_phase_damping",
                        "braket_noise_two_qubit_dephasing",
                        "braket_noise_two_qubit_depolarizing",
                        "braket_unitary_matrix",
                        "braket_result_type_sample",
                        "braket_result_type_expectation",
                        "braket_result_type_variance",
                        "braket_result_type_probability",
                        "braket_result_type_density_matrix",
                    ],
                }
            )
        )
        device_properties.action[DeviceActionType.OPENQASM] = oq3_action
        return device_properties


class DummyAnnealingSimulator(BraketSimulator):
    def run(self, problem: ir.annealing.Problem, *args, **kwargs) -> AnnealingTaskResult:
        return ANNEALING_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        return DeviceCapabilities.parse_obj(
            {
                "service": {
                    "executionWindows": [
                        {
                            "executionDay": "Everyday",
                            "windowStartHour": "11:00",
                            "windowEndHour": "12:00",
                        }
                    ],
                    "shotsRange": [1, 10],
                },
                "action": {
                    "braket.ir.annealing.problem": {
                        "actionType": "braket.ir.annealing.problem",
                        "version": ["1"],
                    }
                },
                "deviceParameters": {},
            }
        )


class DummyRydbergSimulator(BraketSimulator):
    def run(
        self, program: AnalogHamiltonianSimulation, *args, **kwargs
    ) -> AnalogHamiltonianSimulationTaskResult:
        return AHS_RESULT

    @property
    def properties(self) -> DeviceCapabilities:
        properties = {
            "service": {
                "executionWindows": [
                    {
                        "executionDay": "Everyday",
                        "windowStartHour": "00:00",
                        "windowEndHour": "23:59:59",
                    }
                ],
                "shotsRange": [0, 10],
            },
            "action": {"braket.ir.ahs.program": {}},
        }

        RydbergSimulatorDeviceCapabilities = create_model(
            "RydbergSimulatorDeviceCapabilities", **properties
        )

        return RydbergSimulatorDeviceCapabilities.parse_obj(properties)


mock_circuit_entry = Mock()
mock_program_entry = Mock()
mock_jaqcd_entry = Mock()
mock_circuit_dm_entry = Mock()
mock_circuit_entry.load.return_value = DummyCircuitSimulator
mock_program_entry.load.return_value = DummyProgramSimulator
mock_jaqcd_entry.load.return_value = DummyJaqcdSimulator
mock_circuit_dm_entry.load.return_value = DummyProgramDensityMatrixSimulator
local_simulator._simulator_devices = {
    "dummy": mock_circuit_entry,
    "dummy_oq3": mock_program_entry,
    "dummy_jaqcd": mock_jaqcd_entry,
    "dummy_oq3_dm": mock_circuit_dm_entry,
}

mock_ahs_program = AnalogHamiltonianSimulation(
    register=AtomArrangement(), hamiltonian=Hamiltonian()
)


def test_load_from_entry_point():
    sim = LocalSimulator("dummy_oq3")
    task = sim.run(Circuit().h(0).cnot(0, 1), 10)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_gate_model():
    dummy = DummyProgramSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(Circuit().h(0).cnot(0, 1), 10)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_circuit():
    dummy = DummyProgramSimulator()
    theta = FreeParameter("theta")
    task = Circuit().rx(angle=theta, target=0)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [task for _ in range(num_tasks)]
    inputs = [{"theta": i} for i in range(num_tasks)]
    batch = device.run_batch(circuits, inputs=inputs, shots=10)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_with_max_parallel():
    dummy = DummyProgramSimulator()
    task = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [task for _ in range(num_tasks)]
    batch = device.run_batch(circuits, shots=10, max_parallel=2)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_with_annealing_problems():
    dummy = DummyAnnealingSimulator()
    problem = Problem(ProblemType.ISING)
    device = LocalSimulator(dummy)
    num_tasks = 3
    problems = [problem for _ in range(num_tasks)]
    batch = device.run_batch(problems, shots=10)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == AnnealingQuantumTaskResult.from_object(ANNEALING_RESULT)


def test_batch_circuit_without_inputs():
    dummy = DummyProgramSimulator()
    bell = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [bell for _ in range(num_tasks)]
    batch = device.run_batch(circuits, shots=10)
    assert len(batch.results()) == num_tasks
    for x in batch.results():
        assert x == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_circuit_with_unbound_parameters():
    dummy = DummyProgramSimulator()
    device = LocalSimulator(dummy)
    theta = FreeParameter("theta")
    task = Circuit().rx(angle=theta, target=0)
    inputs = {"beta": 0.2}
    cannot_execute_with_unbound = "Cannot execute circuit with unbound parameters: {'theta'}"
    with pytest.raises(ValueError, match=cannot_execute_with_unbound):
        device.run_batch(task, inputs=inputs, shots=10)


def test_batch_circuit_with_single_task():
    dummy = DummyProgramSimulator()
    bell = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    batch = device.run_batch(bell, shots=10)
    assert len(batch.results()) == 1
    assert batch.results()[0] == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_batch_circuit_with_task_and_input_mismatch():
    dummy = DummyProgramSimulator()
    bell = Circuit().h(0).cnot(0, 1)
    device = LocalSimulator(dummy)
    num_tasks = 3
    circuits = [bell for _ in range(num_tasks)]
    inputs = [{} for _ in range(num_tasks - 1)]
    with pytest.raises(ValueError):
        device.run_batch(circuits, inputs=inputs, shots=10)


def test_run_gate_model_inputs():
    dummy = DummyProgramSimulator()
    dummy.run = Mock(return_value=GATE_MODEL_RESULT)
    sim = LocalSimulator(dummy)
    circuit = Circuit().rx(0, FreeParameter("theta"))
    task = sim.run(circuit, inputs={"theta": 2}, shots=10)
    dummy.run.assert_called_with(
        Program(
            source="\n".join(
                (
                    "OPENQASM 3.0;",
                    "input float theta;",
                    "bit[1] b;",
                    "qubit[1] q;",
                    "rx(theta) q[0];",
                    "b[0] = measure q[0];",
                )
            ),
            inputs={"theta": 2},
        ),
        10,
    )
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_program_model_inputs():
    dummy = DummyProgramSimulator()
    dummy.run = Mock(return_value=GATE_MODEL_RESULT)
    sim = LocalSimulator(dummy)
    inputs = {"theta": 2}
    source_string = (
        "OPENQASM 3.0;",
        "input float theta;",
        "bit[1] b;",
        "qubit[1] q;",
        "rx(theta) q[0];",
        "b[0] = measure q[0];",
    )
    program = Program.construct(source="\n".join(source_string), inputs=inputs)
    update_inputs = {"beta": 3}
    task = sim.run(program, inputs=update_inputs, shots=10)
    assert program.inputs == inputs
    program.inputs.update(update_inputs)
    dummy.run.assert_called_with(program, 10)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_jaqcd_only():
    dummy = DummyJaqcdSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(Circuit().h(0).cnot(0, 1), 10)
    dummy.assert_shots(10)
    dummy.assert_qubits(2)
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


def test_run_program_model():
    dummy = DummyProgramSimulator()
    sim = LocalSimulator(dummy)
    task = sim.run(
        Program(
            source="""
qubit[2] q;
bit[2] c;

h q[0];
cnot q[0], q[1];

c = measure q;
"""
        )
    )
    assert task.result() == GateModelQuantumTaskResult.from_object(GATE_MODEL_RESULT)


@pytest.mark.xfail(raises=ValueError)
def test_run_gate_model_value_error():
    dummy = DummyCircuitSimulator()
    sim = LocalSimulator(dummy)
    sim.run(Circuit().h(0).cnot(0, 1))


def test_run_annealing():
    sim = LocalSimulator(DummyAnnealingSimulator())
    task = sim.run(Problem(ProblemType.ISING))
    assert task.result() == AnnealingQuantumTaskResult.from_object(ANNEALING_RESULT)


def test_run_ahs():
    sim = LocalSimulator(DummyRydbergSimulator())
    task = sim.run(mock_ahs_program)
    assert task.result() == AnalogHamiltonianSimulationQuantumTaskResult.from_object(AHS_RESULT)

    task = sim.run(mock_ahs_program.to_ir())
    assert task.result() == AnalogHamiltonianSimulationQuantumTaskResult.from_object(AHS_RESULT)


def test_registered_backends():
    assert LocalSimulator.registered_backends() == {
        "dummy",
        "dummy_oq3",
        "dummy_jaqcd",
        "dummy_oq3_dm",
    }


@pytest.mark.xfail(raises=TypeError)
def test_init_invalid_backend_type():
    LocalSimulator(1234)


@pytest.mark.xfail(raises=ValueError)
def test_init_unregistered_backend():
    LocalSimulator("foo")


@pytest.mark.xfail(raises=NotImplementedError)
def test_run_unsupported_type():
    sim = LocalSimulator(DummyCircuitSimulator())
    sim.run("I'm unsupported")


@pytest.mark.xfail(raises=NotImplementedError)
def test_run_annealing_unsupported():
    sim = LocalSimulator(DummyCircuitSimulator())
    sim.run(Problem(ProblemType.ISING))


@pytest.mark.xfail(raises=NotImplementedError)
def test_run_qubit_gate_unsupported():
    sim = LocalSimulator(DummyAnnealingSimulator())
    sim.run(Circuit().h(0).cnot(0, 1), 1000)


def test_properties():
    dummy = DummyCircuitSimulator()
    sim = LocalSimulator(dummy)
    expected_properties = dummy.properties
    assert sim.properties == expected_properties


@pytest.fixture
def noise_model():
    return (
        NoiseModel()
        .add_noise(Noise.BitFlip(0.05), GateCriteria(Gate.H))
        .add_noise(Noise.TwoQubitDepolarizing(0.10), GateCriteria(Gate.CNot))
    )


@pytest.mark.parametrize("backend", ["dummy_oq3_dm"])
def test_valid_local_device_for_noise_model(backend, noise_model):
    device = LocalSimulator(backend, noise_model=noise_model)
    assert device._noise_model.instructions == [
        NoiseModelInstruction(Noise.BitFlip(0.05), GateCriteria(Gate.H)),
        NoiseModelInstruction(Noise.TwoQubitDepolarizing(0.10), GateCriteria(Gate.CNot)),
    ]


@pytest.mark.parametrize("backend", ["dummy_oq3"])
def test_invalid_local_device_for_noise_model(backend, noise_model):
    with pytest.raises(ValueError):
        _ = LocalSimulator(backend, noise_model=noise_model)


@pytest.mark.parametrize("backend", ["dummy_oq3_dm"])
def test_local_device_with_invalid_noise_model(backend, noise_model):
    with pytest.raises(TypeError):
        _ = LocalSimulator(backend, noise_model=Mock())


@patch.object(DummyProgramDensityMatrixSimulator, "run")
def test_run_with_noise_model(mock_run, noise_model):
    mock_run.return_value = GATE_MODEL_RESULT
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    circuit = Circuit().h(0).cnot(0, 1)
    _ = device.run(circuit, shots=4)

    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[2] b;
        qubit[2] q;
        h q[0];
        #pragma braket noise bit_flip(0.05) q[0]
        cnot q[0], q[1];
        #pragma braket noise two_qubit_depolarizing(0.1) q[0], q[1]
        b[0] = measure q[0];
        b[1] = measure q[1];
        """
    ).strip()

    mock_run.assert_called_with(
        Program(source=expected_circuit, inputs={}),
        4,
    )


@patch.object(LocalSimulator, "_apply_noise_model_to_circuit")
def test_run_batch_with_noise_model(mock_apply, noise_model):
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    circuit = Circuit().h(0).cnot(0, 1)

    mock_apply.return_value = noise_model.apply(circuit)
    _ = device.run_batch([circuit] * 2, shots=4).results()
    assert mock_apply.call_count == 2


@patch.object(DummyProgramDensityMatrixSimulator, "run")
def test_run_noisy_circuit_with_noise_model(mock_run, noise_model):
    mock_run.return_value = GATE_MODEL_RESULT
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    circuit = Circuit().h(0).depolarizing(0, 0.1)
    with warnings.catch_warnings(record=True) as w:
        _ = device.run(circuit, shots=4)

    expected_warning = (
        "The noise model of the device is applied to a circuit that already has noise "
        "instructions."
    )
    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[1] b;
        qubit[1] q;
        h q[0];
        #pragma braket noise bit_flip(0.05) q[0]
        #pragma braket noise depolarizing(0.1) q[0]
        b[0] = measure q[0];
        """
    ).strip()

    mock_run.assert_called_with(
        Program(source=expected_circuit, inputs={}),
        4,
    )
    assert w[-1].message.__str__() == expected_warning


@patch.object(DummyProgramDensityMatrixSimulator, "run")
def test_run_openqasm_with_noise_model(mock_run, noise_model):
    mock_run.return_value = GATE_MODEL_RESULT
    device = LocalSimulator("dummy_oq3_dm", noise_model=noise_model)
    expected_circuit = textwrap.dedent(
        """
        OPENQASM 3.0;
        bit[1] b;
        qubit[1] q;
        h q[0];
        b[0] = measure q[0];
        """
    ).strip()
    expected_warning = (
        "Noise model is only applicable to circuits. The type of the task specification "
        "is Program. The noise model of the device does not apply."
    )
    circuit = Program(source=expected_circuit)
    with warnings.catch_warnings(record=True) as w:
        _ = device.run(circuit, shots=4)

    mock_run.assert_called_with(
        Program(source=expected_circuit, inputs=None),
        4,
    )
    assert w[-1].message.__str__() == expected_warning



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock

import pytest

from braket.circuits import Circuit, Gate, Noise, Observable
from braket.circuits.gates import Unitary
from braket.circuits.noise_model import (
    CircuitInstructionCriteria,
    Criteria,
    GateCriteria,
    NoiseModel,
    ObservableCriteria,
    QubitInitializationCriteria,
    UnitaryGateCriteria,
)
from braket.circuits.noises import BitFlip, Depolarizing, PauliChannel, TwoQubitDepolarizing


def h_unitary():
    return Unitary(Gate.H().to_matrix())


@pytest.fixture
def default_noise_model():
    noise_list = []
    criteria_list = []
    noise_model = NoiseModel()
    for i in range(3):
        noise = Mock(spec=Noise)
        criteria = Mock(spec=CircuitInstructionCriteria)
        criteria.instruction_matches.return_value = i % 2 == 0
        noise_model.add_noise(noise, criteria)
        noise_list.append(noise)
        criteria_list.append(criteria)
    return noise_model, noise_list, criteria_list


def test_simple_add_noise():
    noise_model = NoiseModel()
    assert len(noise_model.instructions) == 0
    mock_noise = Mock(spec=Noise)
    mock_criteria = Mock(spec=CircuitInstructionCriteria)
    noise_model.add_noise(mock_noise, mock_criteria)
    noise_list = noise_model.instructions
    assert len(noise_list) == 1
    assert mock_noise == noise_list[0].noise
    assert mock_criteria == noise_list[0].criteria


def test_insert_noise(default_noise_model):
    noise_model, noise_list, criteria_list = default_noise_model
    listed_noise = noise_model.instructions
    assert len(listed_noise) == 3

    mock_noise = Mock(spec=Noise)
    mock_criteria = Mock(spec=CircuitInstructionCriteria)

    noise_model.insert_noise(1, mock_noise, mock_criteria)

    listed_noise = noise_model.instructions
    assert len(listed_noise) == 4
    assert listed_noise[0].noise == noise_list[0]
    assert listed_noise[0].criteria == criteria_list[0]
    assert listed_noise[1].noise == mock_noise
    assert listed_noise[1].criteria == mock_criteria
    assert listed_noise[2].noise == noise_list[1]
    assert listed_noise[2].criteria == criteria_list[1]
    assert listed_noise[3].noise == noise_list[2]
    assert listed_noise[3].criteria == criteria_list[2]


def test_remove_noise(default_noise_model):
    noise_model, noise_list, _ = default_noise_model

    listed_noise = noise_model.instructions
    assert len(listed_noise) == 3

    noise_model.remove_noise(1)

    listed_noise = noise_model.instructions
    assert len(listed_noise) == 2
    assert listed_noise[0].noise == noise_list[0]
    assert listed_noise[1].noise == noise_list[2]


@pytest.mark.parametrize(
    "gate, qubit, noise, noise_type, expected_length",
    [
        (None, None, None, None, 6),
        (None, None, PauliChannel, None, 2),
        (Gate.H, None, None, None, 3),
        (Gate.CNot, None, None, None, 2),
        (None, 0, None, None, 5),
        # (None, (0, 1), None, None, 1), # TODO: I'm not sure of the best way to fix this.
        (Gate.CNot, (0, 1), None, None, 1),
        (Gate.CNot, (1, 0), None, None, 0),
        (Gate.H, 2, None, None, 1),
        (Gate.H, None, Depolarizing, None, 2),
    ],
)
def test_filter(gate, qubit, noise, noise_type, expected_length):
    noise_model = (
        NoiseModel()
        .add_noise(PauliChannel(0.01, 0.02, 0.03), GateCriteria(Gate.I, [0, 1]))
        .add_noise(Depolarizing(0.04), GateCriteria(Gate.H))
        .add_noise(Depolarizing(0.05), GateCriteria(Gate.CNot, [0, 1]))
        .add_noise(PauliChannel(0.06, 0.07, 0.08), GateCriteria(Gate.H, [0, 1]))
        .add_noise(Depolarizing(0.09), GateCriteria(None, 0))
        .add_noise(Depolarizing(0.10), QubitInitializationCriteria([0, 1]))
    )
    result_model = noise_model.from_filter(qubit=qubit, gate=gate, noise=noise)
    assert len(result_model.instructions) == expected_length


@pytest.mark.parametrize(
    "noise_types, expected_string",
    [
        ([], ""),
        ([Criteria], ""),
        (
            [GateCriteria],
            "Gate Noise:\n  my_noise 0, my_criteria 0\n  my_noise 1, my_criteria 1",
        ),
        (
            [Criteria, GateCriteria, ObservableCriteria, QubitInitializationCriteria],
            "Initialization Noise:\n  my_noise 0, my_criteria 0\n  my_noise 1, my_criteria 1\n"
            "Gate Noise:\n  my_noise 0, my_criteria 0\n  my_noise 1, my_criteria 1\n"
            "Readout Noise:\n  my_noise 0, my_criteria 0\n  my_noise 1, my_criteria 1",
        ),
    ],
)
def test_str_representation(noise_types, expected_string):
    noise_model = NoiseModel()

    for noise_type in noise_types:
        for index in range(2):
            mock_noise = Mock(spec=Noise)
            mock_noise.__str__ = Mock(return_value=f"my_noise {index}")
            mock_criteria = Mock(spec=noise_type)
            mock_criteria.__str__ = Mock(return_value=f"my_criteria {index}")
            noise_model.add_noise(mock_noise, mock_criteria)

    str_representation = noise_model.__str__()
    assert str_representation == expected_string


def test_repr_representation():
    noise_model = NoiseModel()
    mock_noise = Mock(spec=Noise)
    mock_noise.__repr__ = Mock(return_value="n")
    mock_criteria = Mock(spec=Criteria)
    mock_criteria.__repr__ = Mock(return_value="c")
    noise_model.add_noise(mock_noise, mock_criteria)
    str_representation = noise_model.__repr__()
    assert str_representation == "{'instructions': [NoiseModelInstruction(noise=n, criteria=c)]}"


def test_serialization():
    noise_model = (
        NoiseModel()
        .add_noise(PauliChannel(0.01, 0.02, 0.03), GateCriteria(Gate.I, [0, 1]))
        .add_noise(Depolarizing(0.04), GateCriteria(Gate.H))
        .add_noise(Depolarizing(0.05), GateCriteria(Gate.CNot, [0, 1]))
        .add_noise(PauliChannel(0.06, 0.07, 0.08), GateCriteria(Gate.H, [0, 1]))
    )
    serialized_model = noise_model.to_dict()
    deserialized_model = NoiseModel.from_dict(serialized_model)
    assert len(deserialized_model.instructions) == len(noise_model.instructions)
    for index, deserialized_item in enumerate(deserialized_model.instructions):
        assert noise_model.instructions[index].noise == deserialized_item.noise
        assert noise_model.instructions[index].criteria == deserialized_item.criteria
    assert deserialized_model is not None


def test_apply():
    noise_model = (
        NoiseModel()
        .add_noise(PauliChannel(0.01, 0.02, 0.03), GateCriteria(Gate.I, [0, 1]))
        .add_noise(Depolarizing(0.04), GateCriteria(Gate.H))
        .add_noise(TwoQubitDepolarizing(0.05), GateCriteria(Gate.CNot, [0, 1]))
        .add_noise(PauliChannel(0.06, 0.07, 0.08), GateCriteria(Gate.H, [0, 1]))
        .add_noise(Depolarizing(0.10), UnitaryGateCriteria(h_unitary(), 0))
        .add_noise(Depolarizing(0.06), ObservableCriteria(Observable.Z, 0))
        .add_noise(Depolarizing(0.09), QubitInitializationCriteria(0))
    )
    layer1 = Circuit().h(0).cnot(0, 1).sample(Observable.Z(), 0)
    layer2 = Circuit().unitary([0], h_unitary().to_matrix())
    circuit = layer1 + layer2
    noisy_circuit_from_circuit = noise_model.apply(circuit)
    expected_circuit = (
        Circuit()
        .depolarizing(0, 0.09)
        .h(0)
        .depolarizing(0, 0.04)
        .pauli_channel(0, 0.06, 0.07, 0.08)
        .cnot(0, 1)
        .two_qubit_depolarizing(0, 1, 0.05)
        .unitary([0], h_unitary().to_matrix())
        .depolarizing(0, 0.10)
        .apply_readout_noise(Depolarizing(0.06), 0)
        .sample(Observable.Z(), 0)
    )
    assert noisy_circuit_from_circuit == expected_circuit


def test_apply_in_order():
    noise_model = (
        NoiseModel()
        .add_noise(Depolarizing(0.01), GateCriteria(Gate.H))
        .add_noise(Depolarizing(0.02), GateCriteria(Gate.H))
    )
    circuit = Circuit().h(0)
    noisy_circuit = noise_model.apply(circuit)
    expected_circuit = circuit.apply_gate_noise([Depolarizing(0.01), Depolarizing(0.02)])
    assert noisy_circuit == expected_circuit


@pytest.mark.parametrize(
    "noise_model, input_circuit, expected_circuit",
    [
        (
            # model with noise on H(0)
            NoiseModel().add_noise(Depolarizing(0.01), GateCriteria(Gate.H, 0)),
            # input circuit has an H gate on qubit 0
            Circuit().h(0).cnot(0, 1),
            # expected circuit has noise on qubit 0
            Circuit().h(0).depolarizing(0, 0.01).cnot(0, 1),
        ),
        (
            # model with noise on H(0)
            NoiseModel().add_noise(Depolarizing(0.01), GateCriteria(Gate.H, 0)),
            # input circuit has two H gates on qubit 0
            Circuit().h(0).h(0).cnot(0, 1),
            # expected circuit has noise on qubit 0
            Circuit().h(0).depolarizing(0, 0.01).h(0).depolarizing(0, 0.01).cnot(0, 1),
        ),
        (
            # model with noise on all gates, on qubits 0, 1
            NoiseModel().add_noise(Depolarizing(0.01), GateCriteria(None, [0, 1])),
            # input circuit
            Circuit().h(0).h(1).cnot(0, 1),
            # expected circuit has noise on qubit 0
            Circuit().h(0).depolarizing(0, 0.01).h(1).depolarizing(1, 0.01).cnot(0, 1),
        ),
        (
            # model with noise on all gates, on qubits [0, 1]
            NoiseModel().add_noise(Depolarizing(0.01), GateCriteria(None, [[0, 1]])),
            # input circuit
            Circuit().h(0).h(1).cnot(0, 1),
            # expected circuit has noise on the CNot gate
            Circuit().h(0).h(1).cnot(0, 1).depolarizing(0, 0.01).depolarizing(1, 0.01),
        ),
        (
            # model with noise on all gates, on qubits [0, 1]
            NoiseModel().add_noise(TwoQubitDepolarizing(0.01), GateCriteria(None, [[0, 1]])),
            # input circuit
            Circuit().h(0).h(1).cnot(0, 1),
            # expected circuit has noise on the CNot gate
            Circuit().h(0).h(1).cnot(0, 1).two_qubit_depolarizing(0, 1, 0.01),
        ),
        (
            # model with noise on a unitary H(0)
            NoiseModel().add_noise(Depolarizing(0.01), UnitaryGateCriteria(h_unitary(), 0)),
            # input circuit has a unitary H gate on qubit 0
            Circuit().unitary([0], h_unitary().to_matrix()).cnot(0, 1),
            # expected circuit has noise on qubit 0
            Circuit().unitary([0], h_unitary().to_matrix()).depolarizing(0, 0.01).cnot(0, 1),
        ),
    ],
)
def test_gate_noise(noise_model, input_circuit, expected_circuit):
    result_circuit = noise_model.apply(input_circuit)
    assert result_circuit == expected_circuit


@pytest.mark.parametrize(
    "noise_model, input_circuit, expected_circuit",
    [
        (
            # model
            NoiseModel().add_noise(Depolarizing(0.01), QubitInitializationCriteria()),
            # input circuit
            Circuit().h(0).cnot(0, 1),
            # expected circuit has noise on both qubits
            Circuit().depolarizing(0, 0.01).depolarizing(1, 0.01).h(0).cnot(0, 1),
        ),
        (
            # model
            NoiseModel().add_noise(Depolarizing(0.01), QubitInitializationCriteria(range(4))),
            # input circuit
            Circuit().h(0).cnot(0, 1),
            # expected circuit has noise on both qubits
            Circuit().depolarizing(0, 0.01).depolarizing(1, 0.01).h(0).cnot(0, 1),
        ),
        (
            # model only specifies noise on one qubit
            NoiseModel().add_noise(Depolarizing(0.01), QubitInitializationCriteria(1)),
            # input circuit
            Circuit().h(0).cnot(0, 1),
            # expected circuit has noise on one qubit
            Circuit().depolarizing(1, 0.01).h(0).cnot(0, 1),
        ),
        (
            # model only specifies noise on an unrelated qubit
            NoiseModel().add_noise(Depolarizing(0.01), QubitInitializationCriteria(2)),
            # input circuit
            Circuit().h(0).cnot(0, 1),
            # expected circuit has no noise applied
            Circuit().h(0).cnot(0, 1),
        ),
        (
            # model does not specify initialization noise
            NoiseModel().add_noise(Depolarizing(0.01), GateCriteria(Gate.X, 2)),
            # input circuit
            Circuit().h(0).cnot(0, 1),
            # expected circuit has no noise applied
            Circuit().h(0).cnot(0, 1),
        ),
    ],
)
def test_apply_initialization_noise(noise_model, input_circuit, expected_circuit):
    result_circuit = noise_model.apply(input_circuit)
    assert result_circuit == expected_circuit


@pytest.mark.parametrize(
    "noise_model, input_circuit, expected_circuit",
    [
        (
            # model
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(Observable.Z)),
            # input circuit has no explicit observables
            Circuit().h(0).cnot(0, 1),
            # expected circuit has no noise applied
            Circuit().h(0).cnot(0, 1),
        ),
        (
            # model has observable criteria only on one qubit
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(Observable.Z, 0)),
            # input circuit has no explicit observables
            Circuit().h(0).cnot(0, 1),
            # expected circuit has no noise applied
            Circuit().h(0).cnot(0, 1),
        ),
        (
            # model
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(Observable.Z)),
            # input circuit has explicit explicit observables
            Circuit().h(0).cnot(0, 1).sample(Observable.Z(), 0),
            # expected circuit has noise applied
            Circuit().h(0).cnot(0, 1).depolarizing(0, 0.01).sample(Observable.Z(), 0),
        ),
        (
            # model
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(Observable.X, 0)),
            # input circuit doesn't contain observable X
            Circuit().h(0).cnot(0, 1),
            # expected circuit has no change.
            Circuit().h(0).cnot(0, 1),
        ),
        (
            # model
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(Observable.X, 0)),
            # input circuit contains observable X
            Circuit().h(0).cnot(0, 1).sample(Observable.X(), 0),
            # expected circuit noise applied.
            Circuit().h(0).cnot(0, 1).depolarizing(0, 0.01).sample(Observable.X(), 0),
        ),
        (
            # model only has an observable on Z
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(Observable.Z, 0)),
            # input circuit contains observable X
            Circuit().h(0).cnot(0, 1).sample(Observable.X(), 0),
            # expected circuit has no change
            Circuit().h(0).cnot(0, 1).sample(Observable.X(), 0),
        ),
        (
            # model uses qubit criteria
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(None, None)),
            # input circuit doesn't contain observables
            Circuit().h(0).cnot(0, 1),
            # expected circuit has no noise applied
            Circuit().h(0).cnot(0, 1),
        ),
        (
            # model uses qubit criteria on non-related qubits
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(None, [2, 3])),
            # input circuit doesn't contain observables
            Circuit().h(0).cnot(0, 1),
            # expected circuit has no change
            Circuit().h(0).cnot(0, 1),
        ),
        (
            # model uses qubit criteria
            NoiseModel().add_noise(Depolarizing(0.01), ObservableCriteria(None, [0, 1])),
            # input circuit contains observable X
            Circuit().h(0).cnot(0, 1).sample(Observable.X(), 0),
            # expected circuit noise applied.
            Circuit().h(0).cnot(0, 1).depolarizing(0, 0.01).sample(Observable.X(), 0),
        ),
        (
            # model uses observable and qubit criteria
            NoiseModel()
            .add_noise(Depolarizing(0.01), ObservableCriteria(Observable.X, 0))
            .add_noise(Depolarizing(0.02), ObservableCriteria(None, [0, 1])),
            # input circuit contains observable X
            Circuit().h(0).cnot(0, 1).sample(Observable.X(), 0),
            # expected circuit noise applied.
            Circuit()
            .h(0)
            .cnot(0, 1)
            .depolarizing(0, 0.01)
            .depolarizing(0, 0.02)
            .sample(Observable.X(), 0),
        ),
        (
            # model uses observable criteria with any observable/qubit.
            NoiseModel().add_noise(BitFlip(0.01), ObservableCriteria(None, None)),
            # input circuit contains many different types of result types for qubit 0
            Circuit()
            .h(0)
            .cnot(0, 1)
            .probability(target=[0, 1])
            .probability(target=0)
            .expectation(observable=Observable.Z(), target=0)
            .sample(observable=Observable.X(), target=0)
            .variance(observable=Observable.Z(), target=0),
            # expected circuit only applies BitFlip once to qubit 0
            Circuit()
            .h(0)
            .cnot(0, 1)
            .probability(target=[0, 1])
            .probability(target=0)
            .expectation(observable=Observable.Z(), target=0)
            .sample(observable=Observable.X(), target=0)
            .variance(observable=Observable.Z(), target=0)
            .apply_readout_noise(BitFlip(0.01), 0),
        ),
        (
            # model uses observable criteria with any observable/qubit.
            NoiseModel().add_noise(BitFlip(0.01), ObservableCriteria(None, None)),
            # input circuit only has a probability result type
            Circuit().h(0).cnot(0, 1).probability(target=[0, 1]).probability(target=0),
            # expected circuit has no noise applied
            Circuit().h(0).cnot(0, 1).probability(target=[0, 1]).probability(target=0),
        ),
    ],
)
def test_apply_readout_noise(noise_model, input_circuit, expected_circuit):
    result_circuit = noise_model.apply(input_circuit)
    assert result_circuit == expected_circuit


@pytest.mark.xfail(raises=IndexError)
def test_remove_noise_at_invalid_index():
    noise_model = NoiseModel()
    noise_model.remove_noise(index=0)
    assert not "should not get here"


@pytest.mark.xfail(raises=ValueError)
def test_add_invalid_noise():
    noise_model = NoiseModel()
    noise_model.add_noise(Mock(), Mock(spec=Criteria))


@pytest.mark.xfail(raises=ValueError)
def test_add_invalid_criteria():
    noise_model = NoiseModel()
    noise_model.add_noise(Mock(spec=Noise), Mock())


@pytest.mark.xfail(raises=ValueError)
def test_apply_to_circuit_list():
    noise_model = NoiseModel()
    noise_model.add_noise(Mock(), Mock(spec=Criteria))
    noise_model.apply([])



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.ir.blackbird import Program as BlackbirdProgram
from braket.task_result import (
    AdditionalMetadata,
    PhotonicModelTaskResult,
    TaskMetadata,
    XanaduMetadata,
)
from braket.tasks import PhotonicModelQuantumTaskResult


@pytest.fixture
def task_metadata():
    return TaskMetadata(**{"id": "task_arn", "deviceId": "arn1", "shots": 100})


@pytest.fixture
def xanadu_metadata():
    return XanaduMetadata(compiledProgram="DECLARE ro BIT[2];")


@pytest.fixture
def blackbird_program():
    return BlackbirdProgram(source="Vac | q[0]")


@pytest.fixture
def additional_metadata(blackbird_program, xanadu_metadata):
    return AdditionalMetadata(action=blackbird_program, xanaduMetadata=xanadu_metadata)


@pytest.fixture
def measurements():
    return [[[1, 2, 3, 4]], [[4, 3, 2, 1]], [[0, 0, 0, 0]]]


@pytest.fixture
def result_1(measurements, task_metadata, additional_metadata):
    return PhotonicModelTaskResult(
        measurements=measurements,
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
    )


@pytest.fixture
def result_1_str(result_1):
    return result_1.json()


@pytest.fixture
def empty_result(task_metadata, additional_metadata):
    updated_metadata = task_metadata.copy()
    updated_metadata.id = "empty_arn"
    return PhotonicModelTaskResult(
        taskMetadata=updated_metadata, additionalMetadata=additional_metadata
    )


def test_from_object_equals_from_string(result_1, result_1_str):
    assert PhotonicModelQuantumTaskResult.from_object(
        result_1
    ) == PhotonicModelQuantumTaskResult.from_string(result_1_str)


def test_equality(result_1, empty_result):
    quantum_result1 = PhotonicModelQuantumTaskResult.from_object(result_1)
    quantum_empty = PhotonicModelQuantumTaskResult.from_object(empty_result)
    non_result = "not a quantum task result"

    assert quantum_result1 == quantum_result1
    assert quantum_result1 != quantum_empty
    assert quantum_result1 != non_result



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys

import pytest
from botocore.exceptions import ClientError
from test_create_quantum_job import decorator_python_version

from braket.aws import AwsDevice
from braket.circuits import Circuit
from braket.devices import Devices
from braket.jobs import get_job_device_arn, hybrid_job


@pytest.fixture
def reservation_arn(aws_session):
    return (
        f"arn:aws:braket:{aws_session.region}"
        f":{aws_session.account_id}:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
    )


def test_create_task_via_invalid_reservation_arn_on_qpu(reservation_arn):
    circuit = Circuit().h(0)
    device = AwsDevice(Devices.IonQ.Harmony)

    with pytest.raises(ClientError, match="Reservation arn is invalid"):
        device.run(
            circuit,
            shots=10,
            reservation_arn=reservation_arn,
        )


def test_create_task_via_reservation_arn_on_simulator(reservation_arn):
    circuit = Circuit().h(0)
    device = AwsDevice(Devices.Amazon.SV1)

    with pytest.raises(ClientError, match="Braket Direct is not supported for"):
        device.run(
            circuit,
            shots=10,
            reservation_arn=reservation_arn,
        )


@pytest.mark.xfail(
    (sys.version_info.major, sys.version_info.minor) != decorator_python_version(),
    raises=RuntimeError,
    reason="Python version mismatch",
)
def test_create_job_with_decorator_via_invalid_reservation_arn(reservation_arn):
    with pytest.raises(ClientError, match="Reservation arn is invalid"):

        @hybrid_job(
            device=Devices.IonQ.Aria1,
            reservation_arn=reservation_arn,
        )
        def hello_job():
            device = AwsDevice(get_job_device_arn())
            bell = Circuit().h(0).cnot(0, 1)
            task = device.run(bell, shots=10)
            measurements = task.result().measurements
            return measurements

        hello_job()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import math
from typing import Any, Optional

from oqpy import FrameVar as OQFrame
from oqpy.base import OQPyExpression

from braket.pulse.port import Port


class Frame:
    """Frame tracks the frame of reference, when interacting with the qubits, throughout the
    execution of a program. See https://openqasm.com/language/openpulse.html#frames for more
    details.
    """

    def __init__(
        self,
        frame_id: str,
        port: Port,
        frequency: float,
        phase: float = 0,
        is_predefined: bool = False,
        properties: Optional[dict[str, Any]] = None,
    ):
        """Initializes a Frame.

        Args:
            frame_id (str): str identifying a unique frame.
            port (Port): port that this frame is attached to.
            frequency (float): frequency to which this frame should be initialized.
            phase (float): phase to which this frame should be initialized. Defaults to 0.
            is_predefined (bool): bool indicating whether this is a predefined frame on
                the device. Defaults to False.
            properties (Optional[dict[str, Any]]): Dict containing properties of this frame.
                Defaults to None.
        """
        self._frame_id = frame_id
        self.port = port
        self.frequency = frequency
        self.phase = phase
        self.is_predefined = is_predefined
        self.properties = properties

    @property
    def id(self) -> str:
        """Returns a str indicating the frame id."""
        return self._frame_id

    def __eq__(self, other: Frame) -> bool:
        return (
            (
                (self.id == other.id)
                and (self.port == other.port)
                and math.isclose(self.frequency, other.frequency)
                and math.isclose(self.phase, other.phase)
            )
            if isinstance(other, Frame)
            else False
        )

    def _to_oqpy_expression(self) -> OQPyExpression:
        return OQFrame(
            port=self.port._to_oqpy_expression(),
            frequency=self.frequency,
            phase=self.phase,
            name=self.id,
            needs_declaration=not self.is_predefined,
        )
